__all__ = ["PropertyType", "Property"]


class PropertyType(object):
    pass


class Property(object):
    class Select(PropertyType):
        """
        Select Property. The user can select value from list set as options parameter
        """

        def __init__(self, label, options, description=""):
            """

            :param label:
            :param options:
            :param description:
            """
            PropertyType.__init__(self)
            self.label = label
            self.options = options
            self.description = description

    class Number(PropertyType):
        """
        The user can set a number value
        """

        def __init__(
            self, label, configurable=False, default_value=None, unit="", description=""
        ):
            """

            :param label:
            :param configurable:
            :param default_value:
            :param unit:
            :param description:
            """
            PropertyType.__init__(self)
            self.label = label
            self.configurable = configurable
            self.default_value = default_value
            self.description = description

    class Text(PropertyType):
        """
        The user can set a text value
        """

        def __init__(self, label, configurable=False, default_value="", description=""):
            """

            :param label:
            :param configurable:
            :param default_value:
            :param description:
            """
            PropertyType.__init__(self)
            self.label = label
            self.configurable = configurable
            self.default_value = default_value
            self.description = description

    class Actor(PropertyType):
        """
        The user select an actor which is available in the system. The value of this variable will be the actor id
        """

        def __init__(self, label, description=""):
            """

            :param label:
            :param description:
            """
            PropertyType.__init__(self)
            self.label = label
            self.configurable = True
            self.description = description

    class Sensor(PropertyType):
        """
        The user select a sensor which is available in the system. The value of this variable will be the sensor id
        """

        def __init__(self, label, description=""):
            """

            :param label:
            :param description:
            """
            PropertyType.__init__(self)
            self.label = label
            self.configurable = True
            self.description = description

    class Kettle(PropertyType):
        """
        The user select a kettle which is available in the system. The value of this variable will be the kettle id
        """

        def __init__(self, label, description=""):
            """

            :param label:
            :param description:
            """

            PropertyType.__init__(self)
            self.label = label
            self.configurable = True
            self.description = description

    class Fermenter(PropertyType):
        """
        The user select a fermenter which is available in the system. The value of this variable will be the fermenter id
        """

        def __init__(self, label, description=""):
            """

            :param label:
            :param description:
            """

            PropertyType.__init__(self)
            self.label = label
            self.configurable = True
            self.description = description



import asyncio
import base64
import json
import logging
import math
import os
import os.path
import pathlib
import re
import sqlite3
import webbrowser
import xml.etree.ElementTree
from os import listdir
from os.path import isfile, join

import aiohttp
import yaml
from aiohttp import web
from cbpi.api import *
from cbpi.api.base import CBPiBase
from cbpi.api.config import ConfigType
from cbpi.api.dataclasses import (Actor, Kettle, NotificationAction,
                                  NotificationType, Sensor)
from cbpi.controller.kettle_controller import KettleController
from voluptuous.schema_builder import message

from ..api.step import StepMove, StepResult, StepState


class UploadController:

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self.logger = logging.getLogger(__name__)

    async def get_kbh_recipes(self):
        try:
            path = self.cbpi.config_folder.get_upload_file("kbh.db")
            conn = sqlite3.connect(path)
            c = conn.cursor()
            c.execute("SELECT ID, Sudname, Status FROM Sud")
            data = c.fetchall()
            result = []
            for row in data:
                element = {"value": str(row[0]), "label": str(row[1])}
                result.append(element)
            return result
        except:
            return []

    async def get_xml_recipes(self):
        try:
            path = self.cbpi.config_folder.get_upload_file("beer.xml")
            e = xml.etree.ElementTree.parse(path).getroot()
            result = []
            counter = 1
            for idx, val in enumerate(e.findall("RECIPE")):
                element = {"value": str(counter), "label": val.find("NAME").text}
                result.append(element)
                counter += 1
            return result
        except:
            return []

    async def get_json_recipes(self):
        try:
            path = self.cbpi.config_folder.get_upload_file("mmum.json")
            e = json.load(open(path))
            result = []
            result.append({"value": str(1), "label": e["Name"]})
            return result
        except:
            return []

    async def get_brewfather_recipes(self, offset=0):
        limit = 50
        length = self.cbpi.config.get("brewfather_list_length", 50)
        repeat = True
        brewfather = True
        result = []
        self.url = "https://api.brewfather.app/v2/recipes"
        brewfather_user_id = self.cbpi.config.get("brewfather_user_id", None)
        if brewfather_user_id == "" or brewfather_user_id is None:
            brewfather = False

        brewfather_api_key = self.cbpi.config.get("brewfather_api_key", None)
        if brewfather_api_key == "" or brewfather_api_key is None:
            brewfather = False

        if brewfather == True:
            encodedData = base64.b64encode(
                bytes(f"{brewfather_user_id}:{brewfather_api_key}", "ISO-8859-1")
            ).decode("ascii")
            headers = {"Authorization": "Basic %s" % encodedData}
            parameters = {"limit": limit}
            while repeat == True:
                try:
                    async with aiohttp.ClientSession(headers=headers) as bf_session:
                        async with bf_session.get(self.url, params=parameters) as r:
                            if r.status == 429:
                                try:
                                    seconds = int(r.headers["Retry-After"])
                                    minutes = round(seconds / 60)
                                except:
                                    seconds = None
                                if not seconds:
                                    logging.error(
                                        "Too many requests to BF api. Try again later"
                                    )
                                    self.cbpi.notify(
                                        "Error",
                                        "Too many requests to BF api. Try again later",
                                        NotificationType.ERROR,
                                    )
                                else:
                                    logging.error(
                                        f"Too many requests to BF api. Try in {minutes} minutes again."
                                    )
                                    self.cbpi.notify(
                                        "Error",
                                        f"Too many requests to BF api. Try in {minutes} minutes again.",
                                        NotificationType.ERROR,
                                    )
                                repeat = False
                                logging.error(r.headers["Retry-After"])
                            else:
                                bf_recipe_list = await r.json()
                        await bf_session.close()
                except Exception as e:
                    logging.error(e)
                    repeat = False
                try:
                    if bf_recipe_list:
                        for row in bf_recipe_list:
                            recipe_id = row["_id"]
                            name = row["name"]
                            element = {"value": recipe_id, "label": name}
                            result.append(element)
                    else:
                        repeat = False
                except Exception as e:
                    logging.error(e)
                try:
                    if len(bf_recipe_list) != limit:
                        repeat = False
                    else:
                        parameters = {"limit": limit, "start_after": recipe_id}
                except Exception as e:
                    logging.error(e)

        try:
            newlist = sorted(result, key=lambda d: d["label"])
            listlength = len(newlist)
            max = math.floor(listlength / length)
            sortlist = []
            for i in range(0, max + 1):
                sortlist.append({"value": i * length, "label": i * length})
            return newlist, sortlist, length
        except:
            logging.error("Return empty BF recipe list")
            sortlist = [{"value": 0, "label": "0"}]
            return result, sortlist, length

    def get_creation_path(self):
        creation_path = self.cbpi.config.get("RECIPE_CREATION_PATH", "upload")
        path = {"path": "upload"} if creation_path == "" else {"path": creation_path}
        return path

    def allowed_file(self, filename, extension):
        return "." in filename and filename.rsplit(".", 1)[1] in set([extension])

    async def FileUpload(self, data):
        fileData = data["File"]
        filename = fileData.filename
        recipe_file = fileData.file
        content_type = fileData.content_type

        if content_type == "text/xml":
            try:
                beer_xml = recipe_file.read().decode("utf-8", "replace")
                if recipe_file and self.allowed_file(filename, "xml"):
                    self.path = self.cbpi.config_folder.get_upload_file("beer.xml")

                    f = open(self.path, "w")
                    f.write(beer_xml)
                    f.close()
                    self.cbpi.notify(
                        "Success",
                        "XML Recipe {} has been uploaded".format(filename),
                        NotificationType.SUCCESS,
                    )
            except Exception as e:
                self.cbpi.notify(
                    "Error",
                    "XML Recipe upload failed: {}".format(e),
                    NotificationType.ERROR,
                )
                pass

        elif content_type == "application/json":
            try:
                mmum_json = recipe_file.read().decode("utf-8", "replace")
                if recipe_file and self.allowed_file(filename, "json"):
                    self.path = self.cbpi.config_folder.get_upload_file("mmum.json")

                    f = open(self.path, "w")
                    f.write(mmum_json)
                    f.close()
                    self.cbpi.notify(
                        "Success",
                        "JSON Recipe {} has been uploaded".format(filename),
                        NotificationType.SUCCESS,
                    )
            except Exception as e:
                self.cbpi.notify(
                    "Error" "JSON Recipe upload failed: {}".format(e),
                    NotificationType.ERROR,
                )
                pass

        elif content_type == "application/octet-stream":
            try:
                content = recipe_file.read()
                if recipe_file and self.allowed_file(filename, "sqlite"):
                    self.path = self.cbpi.config_folder.get_upload_file("kbh.db")

                    f = open(self.path, "wb")
                    f.write(content)
                    f.close()
                    self.cbpi.notify(
                        "Success",
                        "Kleiner Brauhelfer database has been uploaded",
                        NotificationType.SUCCESS,
                    )

            except:
                self.cbpi.notify(
                    "Error",
                    "Kleiner Brauhelfer database upload failed",
                    NotificationType.ERROR,
                )
                pass
        else:
            self.cbpi.notify(
                "Error", "Wrong content type. Upload failed", NotificationType.ERROR
            )

    # function to create a recipe from the Kleiner Brauhelfer database
    async def kbh_recipe_creation(self, Recipe_ID):
        config = self.get_config_values()
        if self.kettle is not None:
            # check if kbh database is available in the upload folder and connect to it
            self.path = self.cbpi.config_folder.get_upload_file("kbh.db")
            if os.path.exists(self.path) is False:
                self.cbpi.notify(
                    "File Not Found",
                    "Please upload a kbh V2 database file",
                    NotificationType.ERROR,
                )

            try:
                # Get Recipe Nmae
                conn = sqlite3.connect(self.path)
                c = conn.cursor()
                c.execute("SELECT Sudname FROM Sud WHERE ID = ?", (Recipe_ID,))
                row = c.fetchone()
                name = row[0]
                # get MashIn Temp
                mashin_temp = None
                c.execute(
                    "SELECT TempWasser FROM Maischplan WHERE Typ = 0 AND SudID = ?",
                    (Recipe_ID,),
                )
                row = c.fetchone()
                try:
                    if self.cbpi.config.get("TEMP_UNIT", "C") == "C":
                        mashin_temp = str(int(row[0]))
                    else:
                        mashin_temp = str(round(9.0 / 5.0 * int(row[0]) + 32))
                except:
                    pass
                # get the hop addition times
                c.execute(
                    "SELECT Zeit, Name, Vorderwuerze FROM Hopfengaben WHERE Vorderwuerze <> 1 AND SudID = ?",
                    (Recipe_ID,),
                )
                hops = c.fetchall()
                whirlpool = []
                #for hop in hops:
                #    if hop[2] == 5: # whirlpool hops are defined with Vorderwuerze = 5 in the KBH database (Ausschlagen)
                #        whirlpool.append(hop)
                #for whirl in whirlpool:
                #    hops.remove(whirl)
                # get the misc addition times
                
                c.execute(
                    "SELECT Zugabedauer, Name FROM WeitereZutatenGaben WHERE Zeitpunkt = 1 AND SudID = ?",
                    (Recipe_ID,),
                )
                miscs = c.fetchall()
                try:
                    c.execute(
                        "SELECT Zeit, Name FROM Hopfengaben WHERE Vorderwuerze = 1 AND SudID = ?",
                        (Recipe_ID,),
                    )
                    FW_Hops = c.fetchall()
                    FirstWort = self.getFirstWort(FW_Hops, "kbh")
                except:
                    FirstWort = "No"

                # get the boiltime from the database
                c.execute("SELECT Kochdauer FROM Sud WHERE ID = ?", (Recipe_ID,))
                row = c.fetchone()
                BoilTime = str(int(row[0]))

                await self.create_recipe(name)

                # create a mashin step if mashin_temp is available (just heating up to temp and waiting for user input)
                if mashin_temp is not None:
                    step_type = self.mashin if self.mashin != "" else "MashInStep"
                    step_string = {
                        "name": "MashIn",
                        "props": {
                            "AutoMode": self.AutoMode,
                            "Kettle": self.id,
                            "Sensor": self.kettle.sensor,
                            "Temp": mashin_temp,
                            "Timer": "0",
                            "Notification": "Target temperature reached. Please add malt.",
                        },
                        "status_text": "",
                        "status": "I",
                        "type": step_type,
                    }
                    await self.create_step(step_string)

                # if mashin_temp is not available, create a mashin step with the temp of the first mash step if addmashin is set to yes
                step_temp_old = 0
                for row in c.execute(
                    "SELECT Name, TempRast, DauerRast FROM Maischplan WHERE Typ <> 0 AND SudID = ?",
                    (Recipe_ID,),
                ):
                    step_temp = (float(row[1]))
                    if mashin_temp is None and self.addmashin == "Yes":
                        step_type = self.mashin if self.mashin != "" else "MashInStep"
                        step_string = {
                            "name": "MashIn",
                            "props": {
                                "AutoMode": self.AutoMode,
                                "Kettle": self.id,
                                "Sensor": self.kettle.sensor,
                                "Temp": (
                                    str(step_temp)
                                    if self.TEMP_UNIT == "C"
                                    else str(round(9.0 / 5.0 * step_temp + 32,1))
                                ),
                                "Timer": "0",
                                "Notification": "Target temperature reached. Please add malt.",
                            },
                            "status_text": "",
                            "status": "I",
                            "type": step_type,
                        }
                        await self.create_step(step_string)
                    
                    try:
                        if float(step_temp) < float(step_temp_old):
                            step_string = {
                                "name": "Temp reduction!",
                                "props": {
                                    "AutoNext": "No",
                                    "Kettle": self.id,
                                    "Notification": f"Temperature reduction from {step_temp_old}  {self.TEMP_UNIT} to {step_temp} {self.TEMP_UNIT}. Please reduce the temperature manually.",
                                },
                                "status_text": "",
                                "status": "I",
                                "type": "NotificationStep",
                            }

                            await self.create_step(step_string)
                    except Exception as e:
                        logging.error(e)


                    # create the mash steps based on the recipe settings (time and temp)
                    step_type = self.mash if self.mash != "" else "MashStep"
                    step_string = {
                        "name": str(row[0]),
                        "props": {
                            "AutoMode": self.AutoMode,
                            "Kettle": self.id,
                            "Sensor": self.kettle.sensor,
                            "Temp": (
                                str(int(row[1]))
                                if self.TEMP_UNIT == "C"
                                else str(round(9.0 / 5.0 * int(row[1]) + 32))
                            ),
                            "Timer": str(int(row[2])),
                        },
                        "status_text": "",
                        "status": "I",
                        "type": step_type,
                    }
                    await self.create_step(step_string)
                    step_temp_old = step_temp

                # MashOut -> Notification step that sends notification and waits for user input to move to next step (AutoNext=No)
                if self.mashout == "NotificationStep":
                    step_string = {
                        "name": "Lautering",
                        "props": {
                            "AutoNext": "No",
                            "Kettle": self.id,
                            "Notification": "Mash Process completed. Please start lautering and press next to start boil.",
                        },
                        "status_text": "",
                        "status": "I",
                        "type": self.mashout,
                    }
                    await self.create_step(step_string)

                # create a boil step with hop alarms and alarm for first wort hops
                Hops , Whirlpool= self.getBoilAlerts(hops, miscs, "kbh")
                step_type = self.boil if self.boil != "" else "BoilStep"
                step_string = {
                    "name": "Boil Step",
                    "props": {
                        "AutoMode": self.AutoMode,
                        "Kettle": self.boilid,
                        "Sensor": self.boilkettle.sensor,
                        "Temp": int(self.BoilTemp),
                        "Timer": BoilTime,
                        "First_Wort": FirstWort[0],
                        "First_Wort_text": FirstWort[1],
                        "LidAlert": "Yes",
                        "Hop_1": Hops[0][0],
                        "Hop_1_text": Hops[0][1],
                        "Hop_2": Hops[1][0],
                        "Hop_2_text": Hops[1][1],
                        "Hop_3": Hops[2][0],
                        "Hop_3_text": Hops[2][1],
                        "Hop_4": Hops[3][0],
                        "Hop_4_text": Hops[3][1],
                        "Hop_5": Hops[4][0],
                        "Hop_5_text": Hops[4][1],
                        "Hop_6": Hops[5][0],
                        "Hop_6_text": Hops[5][1],
                    },
                    "status_text": "",
                    "status": "I",
                    "type": step_type,
                }

                await self.create_step(step_string)

                # whirlpool hops are added at the end of the boil step
                # and the kettle is cooled down to the whirlpool temperature
                # the whirlpool temperature is set in the recipe and is used to cool down the kettle
                
                if Whirlpool != []:
 
                    step_type = self.cooldown
                    step_name = "CoolDown for Whirlpool Hop"
                    cooldown_sensor = ""
                    step_temp = float(Whirlpool)
                    step_timer = ""

                    if step_type.find("Cooldown") != -1 and self.cooldown != "":
                        cooldown_sensor = (
                                self.boilkettle.sensor
                            )  # fall back to boilkettle sensor if no other sensor is specified
                        step_string = {
                            "name": "Cooldown for Whirlpool Hop",
                            "props": {
                                "Kettle": self.boilid,
                                "Timer": step_timer,
                                "Temp": step_temp,
                                "Sensor": cooldown_sensor,
                                "Actor": self.CoolDownActor,
                            },
                            "status_text": "",
                            "status": "I",
                            "type": step_type,
                        }
                        await self.create_step(step_string)

                    if self.cooldown.find("Cooldown") != -1:
                        notification = "Target Whirlpool temperature reached. Please add Whirlpool hops." 
                        autonext = "No"
                    else: 
                        notification= f"Cool down to {step_temp} {self.TEMP_UNIT}. Then add Whirlpool hops."
                        autonext = "No"
                    step_string = {
                        "name": "Whirlpool Hop",
                        "props": {
                            "AutoNext": autonext,
                            "Kettle": self.id,
                            "Notification": notification,
                        },
                        "status_text": "",
                        "status": "I",
                        "type": "NotificationStep",
                    }
                    await self.create_step(step_string)

                # create a whirlpool step with optional cooldown
                if not whirlpool:
                    await self.create_Whirlpool_Cooldown()
                else:
                    await self.create_Whirlpool_Cooldown(
                        str(abs(whirlpool[0][0]))
                    )  # from kbh this value comes as negative but must be positive

                self.cbpi.notify("KBH Recipe created", name, NotificationType.INFO)

            except Exception as e:
                self.cbpi.notify(
                    "KBH Recipe creation failure: {}".format(e),
                    name,
                    NotificationType.ERROR,
                )
                pass
        else:
            self.cbpi.notify(
                "Recipe Upload",
                "No default Kettle defined. Please specify default Kettle in settings",
                NotificationType.ERROR,
            )

    def getJsonMashin(self, id):
        self.path = self.cbpi.config_folder.get_upload_file("mmum.json")
        e = json.load(open(self.path))
        return float(e["Einmaischtemperatur"])

    # function to create a recipe from a MUMM json recipe file
    async def json_recipe_creation(self, Recipe_ID):
        config = self.get_config_values()
        try:
            if self.kettle is not None:
                # check if  mmum-json file is located in upload folder and load it
                self.path = self.cbpi.config_folder.get_upload_file("mmum.json")
                if os.path.exists(self.path) is False:
                    self.cbpi.notify(
                        "File Not Found",
                        "Please upload a MMuM-JSON File",
                        NotificationType.ERROR,
                    )

                e = json.load(open(self.path))
                logging.info(json.dumps(e, indent=4))
                name = e["Name"]
                boil_time = float(e["Kochzeit_Wuerze"])

                logging.info(name)
                logging.info(boil_time)

                await self.create_recipe(name)

                # get the hop addition times
                hops = []
                firstHops = []
                whirlpool_hops = []

                Hopfenliste=e["Hopfenkochen"]
                for Hopfen in Hopfenliste:
                    #logging.error(Hopfen)
                    if Hopfen["Typ"] == "Standard":
                        hops.append({"name": Hopfen["Sorte"], "time": Hopfen["Zeit"]})
                    if Hopfen["Typ"] == "Vorderwuerze":
                        firstHops.append({"name": Hopfen["Sorte"]})
                    if Hopfen["Typ"] == "Whirlpool":
                        whirlpool_hops.append({"name": Hopfen["Sorte"]})
 
                #logging.error(hops)
                #logging.error(firstHops)
                #logging.error(whirlpool_hops)

                FirstWort = self.getFirstWort(firstHops, "json")
                #logging.error(FirstWort)
                miscs = []
                try:
                    weitere_zutaten=e["Gewuerze_etc"]
                    for zutat in weitere_zutaten:
                        miscs.append({"name": zutat["Name"], "time": zutat["Kochzeit"]})
                #logging.error(miscs)
                except:
                    pass
                # Mash Steps -> first step is different as it heats up to defined temp and stops with notification to add malt
                # AutoMode is yes to start and stop automatic mode or each step
                MashIn_Flag = True
                step_kettle = self.id
                last_step_temp = 0
                step_temp_old = 0
                logging.info(
                    step_kettle
                )  ###################################################
                for row in self.getSteps(Recipe_ID, "json"):
                    step_name = str(row.get("name"))
                    step_timer = str(int(row.get("timer")))
                    step_temp = str(int(row.get("temp")))
                    last_step_temp = step_temp
                    sensor = self.kettle.sensor
                    if MashIn_Flag == True:
                        if row.get("timer") == 0:
                            step_type = (
                                self.mashin if self.mashin != "" else "MashInStep"
                            )
                            Notification = (
                                "Target temperature reached. Please add malt."
                            )
                            MashIn_Flag = False
                            if step_name is None or step_name == "":
                                step_name = "MashIn"
                        elif self.addmashin == "Yes":
                            step_type = (
                                self.mashin if self.mashin != "" else "MashInStep"
                            )
                            Notification = (
                                "Target temperature reached. Please add malt."
                            )
                            MashIn_Flag = False
                            step_string = {
                                "name": "MashIn",
                                "props": {
                                    "AutoMode": self.AutoMode,
                                    "Kettle": self.id,
                                    "Sensor": self.kettle.sensor,
                                    "Temp": self.getJsonMashin(Recipe_ID),
                                    "Timer": 0,
                                    "Notification": Notification,
                                },
                                "status_text": "",
                                "status": "I",
                                "type": step_type,
                            }
                            await self.create_step(step_string)
                            logging.info(
                                step_kettle
                            )  ###################################################

                            step_type = self.mash if self.mash != "" else "MashStep"
                            Notification = ""
                        else:
                            step_type = self.mash if self.mash != "" else "MashStep"
                            Notification = ""

                    else:
                        step_type = self.mash if self.mash != "" else "MashStep"
                        Notification = ""

                    try:
                        if float(step_temp) < float(step_temp_old):
                            step_string = {
                                "name": "Temp reduction!",
                                "props": {
                                    "AutoNext": "No",
                                    "Kettle": self.id,
                                    "Notification": f"Temperature reduction from {step_temp_old}  {self.TEMP_UNIT} to {step_temp} {self.TEMP_UNIT}. Please reduce the temperature manually.",
                                },
                                "status_text": "",
                                "status": "I",
                                "type": "NotificationStep",
                            }

                            await self.create_step(step_string)
                    except Exception as e:
                        logging.error(e)

                    step_string = {
                        "name": step_name,
                        "props": {
                            "AutoMode": self.AutoMode,
                            "Kettle": self.id,
                            "Sensor": self.kettle.sensor,
                            "Temp": step_temp,
                            "Timer": step_timer,
                            "Notification": Notification,
                        },
                        "status_text": "",
                        "status": "I",
                        "type": step_type,
                    }

                    await self.create_step(step_string)
                    step_temp_old = step_temp
                # MashOut -> mashStep to reach mashout-temp for 1 min
                if last_step_temp != e["Abmaischtemperatur"]:
                    step_string = {
                        "name": "MashOut",
                        "props": {
                            "AutoMode": self.AutoMode,
                            "Kettle": self.id,
                            "Sensor": self.kettle.sensor,
                            "Temp": e["Abmaischtemperatur"],
                            "Timer": 1,
                            "Notification": "",
                        },
                        "status_text": "",
                        "status": "I",
                        "type": "MashStep",
                    }

                    await self.create_step(step_string)
                # Lautering -> Simple step that sends notification and waits for user input to move to next step (AutoNext=No)
                if self.mashout == "NotificationStep":
                    step_string = {
                        "name": "Lautering",
                        "props": {
                            "AutoNext": "No",
                            "Kettle": self.id,
                            "Notification": "Mash Process completed. Please start lautering and press next to start boil.",
                        },
                        "status_text": "",
                        "status": "I",
                        "type": self.mashout,
                    }
                    await self.create_step(step_string)

                # Measure Original Gravity -> Simple step that sends notification
                step_string = {
                    "name": "Measure Original Gravity",
                    "props": {
                        "AutoNext": "No",
                        "Kettle": self.id,
                        "Notification": "What is the original gravity of the beer wort?",
                    },
                    "status_text": "",
                    "status": "I",
                    "type": "NotificationStep",
                }
                await self.create_step(step_string)

                # Boil step including hop alarms and alarm for first wort hops -> Automode is set tu yes
                Hops, Whirlpool = self.getBoilAlerts(hops, miscs, "json")
                step_kettle = self.boilid
                step_type = self.boil if self.boil != "" else "BoilStep"
                step_time = str(int(boil_time))
                step_temp = self.BoilTemp
                sensor = self.boilkettle.sensor
                LidAlert = "Yes"

                logging.info(
                    step_temp
                )  ###################################################

                step_string = {
                    "name": "Boil Step",
                    "props": {
                        "AutoMode": self.AutoMode,
                        "Kettle": step_kettle,
                        "Sensor": sensor,
                        "Temp": step_temp,
                        "Timer": step_time,
                        "First_Wort": FirstWort[0],
                        "First_Wort_text": FirstWort[1],
                        "LidAlert": LidAlert,
                        "Hop_1": Hops[0][0],
                        "Hop_1_text": Hops[0][1],
                        "Hop_2": Hops[1][0],
                        "Hop_2_text": Hops[1][1],
                        "Hop_3": Hops[2][0],
                        "Hop_3_text": Hops[2][1],
                        "Hop_4": Hops[3][0],
                        "Hop_4_text": Hops[3][1],
                        "Hop_5": Hops[4][0],
                        "Hop_5_text": Hops[4][1],
                        "Hop_6": Hops[5][0],
                        "Hop_6_text": Hops[5][1],
                    },
                    "status_text": "",
                    "status": "I",
                    "type": step_type,
                }

                await self.create_step(step_string)

                # Measure Original Gravity -> Simple step that sends notification
                step_string = {
                    "name": "Measure Original Gravity",
                    "props": {
                        "AutoNext": "No",
                        "Kettle": self.id,
                        "Notification": "What is the original gravity of the beer wort?",
                    },
                    "status_text": "",
                    "status": "I",
                    "type": "NotificationStep",
                }
                await self.create_step(step_string)

                                # whirlpool hops are added at the end of the boil step
                # and the kettle is cooled down to the whirlpool temperature
                # the whirlpool temperature is set in the recipe and is used to cool down the kettle
                
                if whirlpool_hops != []:

                    step_temp = 80 if self.TEMP_UNIT == "C" else 176

                    step_type = self.cooldown
                    step_name = "CoolDown for Whirlpool Hop"
                    cooldown_sensor = ""
                    step_timer = ""

                    if step_type.find("Cooldown") != -1 and self.cooldown != "":
                        cooldown_sensor = (
                                self.boilkettle.sensor
                            )  # fall back to boilkettle sensor if no other sensor is specified
                        step_string = {
                            "name": "Cooldown for Whirlpool Hop",
                            "props": {
                                "Kettle": self.boilid,
                                "Timer": step_timer,
                                "Temp": step_temp,
                                "Sensor": cooldown_sensor,
                                "Actor": self.CoolDownActor,
                            },
                            "status_text": "",
                            "status": "I",
                            "type": step_type,
                        }
                        await self.create_step(step_string)

                    if self.cooldown.find("Cooldown") != -1:
                        notification = "Target Whirlpool temperature reached. Please add Whirlpool hops." 
                        autonext = "No"
                    else: 
                        notification= f"Cool down to {step_temp} {self.TEMP_UNIT}. Then add Whirlpool hops."
                        autonext = "No"
                    step_string = {
                        "name": "Whirlpool Hop",
                        "props": {
                            "AutoNext": autonext,
                            "Kettle": self.id,
                            "Notification": notification,
                        },
                        "status_text": "",
                        "status": "I",
                        "type": "NotificationStep",
                    }
                    await self.create_step(step_string)



                await self.create_Whirlpool_Cooldown()

                self.cbpi.notify(
                    "MMuM-JSON Recipe created ", name, NotificationType.INFO
                )
            else:
                self.cbpi.notify(
                    "Recipe Upload",
                    "No default Kettle defined. Please specify default Kettle in settings",
                    NotificationType.ERROR,
                )
        except Exception as e:
            self.cbpi.notify(
                "MMuM-JSON Recipe creation failure: {}".format(e),
                name,
                NotificationType.ERROR,
            )
            self.logger.error(e)

    async def xml_recipe_creation(self, Recipe_ID):
        config = self.get_config_values()
        try:
            if self.kettle is not None:
                # load beerxml file located in upload folder
                self.path = self.cbpi.config_folder.get_upload_file("beer.xml")
                if os.path.exists(self.path) is False:
                    self.cbpi.notify(
                        "File Not Found",
                        "Please upload a Beer.xml File",
                        NotificationType.ERROR,
                    )

                e = xml.etree.ElementTree.parse(self.path).getroot()
                recipe = e.find("./RECIPE[%s]" % (str(Recipe_ID)))
                hops = recipe.findall("./HOPS/HOP")
                miscs = recipe.findall('MISCS/MISC[USE="Boil"]')
                name = e.find("./RECIPE[%s]/NAME" % (str(Recipe_ID))).text
                boil_time = float(
                    e.find("./RECIPE[%s]/BOIL_TIME" % (str(Recipe_ID))).text
                )
                FirstWort = self.getFirstWort(hops, "xml")

                await self.create_recipe(name)
                # Mash Steps -> first step is different as it heats up to defined temp and stops with notification to add malt
                # AutoMode is yes to start and stop automatic mode or each step
                MashIn_Flag = True
                step_kettle = self.id
                step_temp_old = 0
                for row in self.getSteps(Recipe_ID, "xml"):
                    step_name = str(row.get("name"))
                    step_timer = str(int(row.get("timer")))
                    step_temp = str(int(row.get("temp")))
                    sensor = self.kettle.sensor
                    if MashIn_Flag == True:
                        if row.get("timer") == 0:
                            step_type = (
                                self.mashin if self.mashin != "" else "MashInStep"
                            )
                            Notification = (
                                "Target temperature reached. Please add malt."
                            )
                            MashIn_Flag = False
                            if step_name is None or step_name == "":
                                step_name = "MashIn"
                        elif self.addmashin == "Yes":
                            step_type = (
                                self.mashin if self.mashin != "" else "MashInStep"
                            )
                            Notification = (
                                "Target temperature reached. Please add malt."
                            )
                            MashIn_Flag = False
                            step_string = {
                                "name": "MashIn",
                                "props": {
                                    "AutoMode": self.AutoMode,
                                    "Kettle": self.id,
                                    "Sensor": self.kettle.sensor,
                                    "Temp": step_temp,
                                    "Timer": 0,
                                    "Notification": Notification,
                                },
                                "status_text": "",
                                "status": "I",
                                "type": step_type,
                            }
                            await self.create_step(step_string)

                            step_type = self.mash if self.mash != "" else "MashStep"
                            Notification = ""
                        else:
                            step_type = self.mash if self.mash != "" else "MashStep"
                            Notification = ""

                    else:
                        step_type = self.mash if self.mash != "" else "MashStep"
                        Notification = ""

                    try:
                        if float(step_temp) < float(step_temp_old):
                            step_string = {
                                "name": "Temp reduction!",
                                "props": {
                                    "AutoNext": "No",
                                    "Kettle": self.id,
                                    "Notification": f"Temperature reduction from {step_temp_old}  {self.TEMP_UNIT} to {step_temp} {self.TEMP_UNIT}. Please reduce the temperature manually.",
                                },
                                "status_text": "",
                                "status": "I",
                                "type": "NotificationStep",
                            }

                            await self.create_step(step_string)
                    except Exception as e:
                        logging.error(e)

                    step_string = {
                        "name": step_name,
                        "props": {
                            "AutoMode": self.AutoMode,
                            "Kettle": self.id,
                            "Sensor": self.kettle.sensor,
                            "Temp": step_temp,
                            "Timer": step_timer,
                            "Notification": Notification,
                        },
                        "status_text": "",
                        "status": "I",
                        "type": step_type,
                    }

                    await self.create_step(step_string)
                    step_temp_old = step_temp

                # MashOut -> Simple step that sends notification and waits for user input to move to next step (AutoNext=No)
                if self.mashout == "NotificationStep":
                    step_string = {
                        "name": "Lautering",
                        "props": {
                            "AutoNext": "No",
                            "Kettle": self.id,
                            "Notification": "Mash Process completed. Please start lautering and press next to start boil.",
                        },
                        "status_text": "",
                        "status": "I",
                        "type": self.mashout,
                    }
                    await self.create_step(step_string)

                # Boil step including hop alarms and alarm for first wort hops -> Automode is set tu yes
                Hops , Whirlpool = self.getBoilAlerts(hops, miscs, "xml")
                step_kettle = self.boilid
                step_type = self.boil if self.boil != "" else "BoilStep"
                step_time = str(int(boil_time))
                step_temp = self.BoilTemp
                sensor = self.boilkettle.sensor
                LidAlert = "Yes"

                step_string = {
                    "name": "Boil Step",
                    "props": {
                        "AutoMode": self.AutoMode,
                        "Kettle": step_kettle,
                        "Sensor": sensor,
                        "Temp": step_temp,
                        "Timer": step_time,
                        "First_Wort": FirstWort[0],
                        "First_Wort_text": FirstWort[1],
                        "LidAlert": LidAlert,
                        "Hop_1": Hops[0][0],
                        "Hop_1_text": Hops[0][1],
                        "Hop_2": Hops[1][0],
                        "Hop_2_text": Hops[1][1],
                        "Hop_3": Hops[2][0],
                        "Hop_3_text": Hops[2][1],
                        "Hop_4": Hops[3][0],
                        "Hop_4_text": Hops[3][1],
                        "Hop_5": Hops[4][0],
                        "Hop_5_text": Hops[4][1],
                        "Hop_6": Hops[5][0],
                        "Hop_6_text": Hops[5][1],
                    },
                    "status_text": "",
                    "status": "I",
                    "type": step_type,
                }

                await self.create_step(step_string)

                # whirlpool hops are added at the end of the boil step
                # and the kettle is cooled down to the whirlpool temperature
                # the whirlpool temperature is set in the recipe and is used to cool down the kettle
                
                if Whirlpool != []:
                    logging.info(
                        "Whirlpool Temp: {}".format(Whirlpool)
                    )

                    step_type = self.cooldown
                    step_name = "CoolDown for Whirlpool Hop"
                    cooldown_sensor = ""
                    step_temp = float(Whirlpool)
                    step_timer = ""

                    if step_type.find("Cooldown") != -1 and self.cooldown != "":
                        cooldown_sensor = (
                                self.boilkettle.sensor
                            )  # fall back to boilkettle sensor if no other sensor is specified
                        step_string = {
                            "name": "Cooldown for Whirlpool Hop",
                            "props": {
                                "Kettle": self.boilid,
                                "Timer": step_timer,
                                "Temp": step_temp,
                                "Sensor": cooldown_sensor,
                                "Actor": self.CoolDownActor,
                            },
                            "status_text": "",
                            "status": "I",
                            "type": step_type,
                        }
                        await self.create_step(step_string)

                    if self.cooldown.find("Cooldown") != -1:
                        notification = "Target Whirlpool temperature reached. Please add Whirlpool hops." 
                        autonext = "No"
                    else: 
                        notification= f"Cool down to {step_temp} {self.TEMP_UNIT}. Then add Whirlpool hops."
                        autonext = "No"
                    step_string = {
                        "name": "Whirlpool Hop",
                        "props": {
                            "AutoNext": autonext,
                            "Kettle": self.id,
                            "Notification": notification,
                        },
                        "status_text": "",
                        "status": "I",
                        "type": "NotificationStep",
                    }
                    await self.create_step(step_string)

                await self.create_Whirlpool_Cooldown()

                self.cbpi.notify("BeerXML Recipe created ", name, NotificationType.INFO)
            else:
                self.cbpi.notify(
                    "Recipe Upload",
                    "No default Kettle defined. Please specify default Kettle in settings",
                    NotificationType.ERROR,
                )
        except Exception as e:
            self.cbpi.notify(
                "BeerXML Recipe creation failure: {}".format(e),
                name,
                NotificationType.ERROR,
            )
            logging.error(e)
            pass

    # XML functions to retrieve xml repice parameters (if multiple recipes are stored in one xml file, id could be used)
    def getSteps(self, id, recipe_type):
        steps = []
        if recipe_type == "xml":
            e = xml.etree.ElementTree.parse(self.path).getroot()
            for e in e.findall("./RECIPE[%s]/MASH/MASH_STEPS/MASH_STEP" % (str(id))):
                if self.cbpi.config.get("TEMP_UNIT", "C") == "C":
                    temp = float(e.find("STEP_TEMP").text)
                else:
                    temp = round(9.0 / 5.0 * float(e.find("STEP_TEMP").text) + 32, 2)
                steps.append(
                    {
                        "name": e.find("NAME").text,
                        "temp": temp,
                        "timer": float(e.find("STEP_TIME").text),
                    }
                )
        elif recipe_type == "json":
            self.path = self.cbpi.config_folder.get_upload_file("mmum.json")
            e = json.load(open(self.path))
            Rasten= e["Rasten"]
            idx = 1
            for Rast in Rasten:
                if self.cbpi.config.get("TEMP_UNIT", "C") == "C":
                    temp = float(Rast["Temperatur"])
                else:
                    temp = round(
                        9.0 / 5.0 * float(Rast["Temperatur"])
                        + 32,
                        2,
                    )

                time= float(Rast["Zeit"])
                steps.append(
                    {
                        "name": "Rast {}".format(idx),
                        "temp": temp,
                        "timer": time,
                    }
                )
                idx += 1
            logging.info(steps)
        return steps

    async def bf_recipe_creation(self, Recipe_ID):
        config = self.get_config_values()

        if self.kettle is not None:

            brewfather = True
            result = []
            self.bf_url = "https://api.brewfather.app/v2/recipes/" + Recipe_ID
            brewfather_user_id = self.cbpi.config.get("brewfather_user_id", None)
            if brewfather_user_id == "" or brewfather_user_id is None:
                brewfather = False

            brewfather_api_key = self.cbpi.config.get("brewfather_api_key", None)
            if brewfather_api_key == "" or brewfather_api_key is None:
                brewfather = False

            if brewfather == True:
                encodedData = base64.b64encode(
                    bytes(f"{brewfather_user_id}:{brewfather_api_key}", "ISO-8859-1")
                ).decode("ascii")
                headers = {"Authorization": "Basic %s" % encodedData}
                bf_recipe = ""

                async with aiohttp.ClientSession(headers=headers) as bf_session:
                    async with bf_session.get(self.bf_url) as r:
                        bf_recipe = await r.json()
                    await bf_session.close()

            if bf_recipe != "":
                try:
                    StrikeTemp = bf_recipe["data"]["strikeTemp"]
                except:
                    StrikeTemp = None
                # BF is sending all Temeprature values in °C. If system is running in F, values need to be converted
                if StrikeTemp is not None and self.TEMP_UNIT != "C":
                    StrikeTemp = round((9.0 / 5.0 * float(StrikeTemp) + 32))

                RecipeName = bf_recipe["name"]
                BoilTime = bf_recipe["boilTime"]
                mash_steps = bf_recipe["mash"]["steps"]
                hops = bf_recipe["hops"]

                try:
                    miscs = bf_recipe["miscs"]
                except:
                    miscs = None

                try:
                    fermentation_steps = bf_recipe["fermentation"]["steps"]
                except:
                    fermentation_steps = None

                if fermentation_steps is not None:
                    try:
                        step = fermentation_steps[0]
                        self.fermentation_step_temp = int(step["stepTemp"])
                    except:
                        self.fermentation_step_temp = None

                if self.fermentation_step_temp is not None and self.TEMP_UNIT != "C":
                    self.fermentation_step_temp = round(
                        (9.0 / 5.0 * float(self.fermentation_step_temp) + 32)
                    )

                FirstWort = self.getFirstWort(hops, "bf")

                await self.create_recipe(RecipeName)

                # Mash Steps -> first step is different as it heats up to defined temp and stops with notification to add malt
                # AutoMode is yes to start and stop automatic mode or each step
                MashIn_Flag = True
                step_kettle = self.id
                step_temp_old = 0
                for step in mash_steps:
                    try:
                        step_name = step["name"]
                        if step_name == "":
                            step_name = "MashStep"
                    except:
                        step_name = "MashStep"

                    step_timer = str(int(step["stepTime"]))

                    if self.TEMP_UNIT == "C":
                        step_temp = str(int(step["stepTemp"]))
                    else:
                        step_temp = str(round((9.0 / 5.0 * int(step["stepTemp"]) + 32)))

                    sensor = self.kettle.sensor
                    if MashIn_Flag == True:

                        if int(step_timer) == 0:
                            step_type = (
                                self.mashin if self.mashin != "" else "MashInStep"
                            )
                            Notification = (
                                "Target temperature reached. Please add malt."
                            )
                            MashIn_Flag = False

                        elif self.addmashin == "Yes":
                            mashin_temp = (
                                str(round(StrikeTemp))
                                if StrikeTemp is not None
                                else step_temp
                            )
                            step_type = (
                                self.mashin if self.mashin != "" else "MashInStep"
                            )
                            Notification = (
                                "Target temperature reached. Please add malt."
                            )
                            MashIn_Flag = False
                            step_string = {
                                "name": "MashIn",
                                "props": {
                                    "AutoMode": self.AutoMode,
                                    "Kettle": self.id,
                                    "Sensor": self.kettle.sensor,
                                    "Temp": mashin_temp,
                                    "Timer": 0,
                                    "Notification": Notification,
                                },
                                "status_text": "",
                                "status": "I",
                                "type": step_type,
                            }
                            await self.create_step(step_string)

                            step_type = self.mash if self.mash != "" else "MashStep"
                            Notification = ""
                        else:
                            step_type = self.mash if self.mash != "" else "MashStep"
                            Notification = ""

                    else:
                        step_type = self.mash if self.mash != "" else "MashStep"
                        Notification = ""
                    try:
                        if float(step_temp) < float(step_temp_old):
                            step_string = {
                                "name": "Temp reduction!",
                                "props": {
                                    "AutoNext": "No",
                                    "Kettle": self.id,
                                    "Notification": f"Temperature reduction from {step_temp_old}  {self.TEMP_UNIT} to {step_temp} {self.TEMP_UNIT}. Please reduce the temperature manually.",
                                },
                                "status_text": "",
                                "status": "I",
                                "type": "NotificationStep",
                            }

                            await self.create_step(step_string)
                    except Exception as e:
                        logging.error(e)

                    step_string = {
                        "name": step_name,
                        "props": {
                            "AutoMode": self.AutoMode,
                            "Kettle": self.id,
                            "Sensor": self.kettle.sensor,
                            "Temp": step_temp,
                            "Timer": step_timer,
                            "Notification": Notification,
                        },
                        "status_text": "",
                        "status": "I",
                        "type": step_type,
                    }

                    await self.create_step(step_string)
                    step_temp_old = step_temp


                # MashOut -> Simple step that sends notification and waits for user input to move to next step (AutoNext=No)

                if self.mashout == "NotificationStep":
                    step_string = {
                        "name": "Lautering",
                        "props": {
                            "AutoNext": "No",
                            "Kettle": self.id,
                            "Notification": "Mash Process completed. Please start lautering and press next to start boil.",
                        },
                        "status_text": "",
                        "status": "I",
                        "type": self.mashout,
                    }
                await self.create_step(step_string)

                # Boil step including hop alarms and alarm for first wort hops -> Automode is set tu yes
                Hops, Whirlpool = self.getBoilAlerts(hops, miscs, "bf")

                step_kettle = self.boilid
                step_time = str(int(BoilTime))
                step_type = self.boil if self.boil != "" else "BoilStep"
                step_temp = self.BoilTemp
                sensor = self.boilkettle.sensor
                LidAlert = "Yes"

                step_string = {
                    "name": "Boil Step",
                    "props": {
                        "AutoMode": self.AutoMode,
                        "Kettle": step_kettle,
                        "Sensor": sensor,
                        "Temp": step_temp,
                        "Timer": step_time,
                        "First_Wort": FirstWort[0],
                        "First_Wort_text": FirstWort[1],
                        "LidAlert": LidAlert,
                        "Hop_1": Hops[0][0],
                        "Hop_1_text": Hops[0][1],
                        "Hop_2": Hops[1][0],
                        "Hop_2_text": Hops[1][1],
                        "Hop_3": Hops[2][0],
                        "Hop_3_text": Hops[2][1],
                        "Hop_4": Hops[3][0],
                        "Hop_4_text": Hops[3][1],
                        "Hop_5": Hops[4][0],
                        "Hop_5_text": Hops[4][1],
                        "Hop_6": Hops[5][0],
                        "Hop_6_text": Hops[5][1],
                    },
                    "status_text": "",
                    "status": "I",
                    "type": step_type,
                }

                await self.create_step(step_string)

                # whirlpool hops are added at the end of the boil step
                # and the kettle is cooled down to the whirlpool temperature
                # the whirlpool temperature is set in the recipe and is used to cool down the kettle
                
                if Whirlpool != []:
                    logging.info(
                        "Whirlpool Temp: {}".format(Whirlpool)
                    )

                    step_type = self.cooldown
                    step_name = "CoolDown for Whirlpool Hop"
                    cooldown_sensor = ""
                    step_temp = float(Whirlpool)
                    step_timer = ""

                    if step_type.find("Cooldown") != -1 and self.cooldown != "":
                        cooldown_sensor = (
                                self.boilkettle.sensor
                            )  # fall back to boilkettle sensor if no other sensor is specified
                        step_string = {
                            "name": "Cooldown for Whirlpool Hop",
                            "props": {
                                "Kettle": self.boilid,
                                "Timer": step_timer,
                                "Temp": step_temp,
                                "Sensor": cooldown_sensor,
                                "Actor": self.CoolDownActor,
                            },
                            "status_text": "",
                            "status": "I",
                            "type": step_type,
                        }
                        await self.create_step(step_string)

                    if self.cooldown.find("Cooldown") != -1:
                        notification = "Target Whirlpool temperature reached. Please add Whirlpool hops." 
                        autonext = "No"
                    else: 
                        notification= f"Cool down to {step_temp} {self.TEMP_UNIT}. Then add Whirlpool hops."
                        autonext = "No"
                    step_string = {
                        "name": "Whirlpool Hop",
                        "props": {
                            "AutoNext": autonext,
                            "Kettle": self.id,
                            "Notification": notification,
                        },
                        "status_text": "",
                        "status": "I",
                        "type": "NotificationStep",
                    }
                    await self.create_step(step_string)


                await self.create_Whirlpool_Cooldown()

                self.cbpi.notify(
                    "Brewfather App Recipe created: ", RecipeName, NotificationType.INFO
                )
        else:
            self.cbpi.notify(
                "Recipe Upload",
                "No default Kettle defined. Please specify default Kettle in settings",
                NotificationType.ERROR,
            )

    def getBoilAlerts(self, hops, miscs, recipe_type):
        alerts = []
        whirlpool = []
        for hop in hops:
            if recipe_type == "xml":
                use = hop.find("USE").text
                ## Hops which are not used in the boil step should not cause alerts
                if use == "Boil":
                    alerts.append([float(hop.find("TIME").text), hop.find("NAME").text])
                elif use == "Aroma":
                    try:
                        if self.TEMP_UNIT == "C":
                            temp = float(hop.find("TEMP").text)
                        else:
                            temp = round(9.0 / 5.0 * float(hop.find("TEMP").text) + 32, 2)
                    except:
                        temp = 80 if self.TEMP_UNIT == "C" else 176
                    whirlpool.append([temp, hop.find("NAME").text])

            elif recipe_type == "bf":
                use = hop["use"]
                if use == "Boil":
                    alerts.append([float(hop["time"]), hop["name"]])  ## TODO: Testing
                elif use == "Aroma":
                    try:
                        if self.TEMP_UNIT == "C":
                            temp = float(hop["temp"])
                        else:
                            temp = round(9.0 / 5.0 * float(hop["temp"]) + 32, 2)
                    except:
                        temp = 80 if self.TEMP_UNIT == "C" else 176
                    whirlpool.append([temp, hop["name"]])

            elif recipe_type == "kbh":
                if hop[2] != 5:
                    alerts.append([float(hop[0]), hop[1]])
                if hop[2] == 5:
                    temp = 80 if self.TEMP_UNIT == "C" else 176
                    whirlpool.append([temp, hop[1]])

            elif recipe_type == "json":
                alerts.append([float(hop["time"]), hop["name"]])



        ## There might also be miscelaneous additions during boild time
        if miscs is not None:
            for misc in miscs:
                if recipe_type == "xml":
                    alerts.append(
                        [float(misc.find("TIME").text), misc.find("NAME").text]
                    )
                elif recipe_type == "bf":
                    use = misc["use"]
                    if use != "Aroma" and use != "Boil":
                        continue
                    alerts.append([float(misc["time"]), misc["name"]])  ## TODO: Testing
                elif recipe_type == "kbh":
                    alerts.append([float(misc[0]), misc[1]])
                elif recipe_type == "json":
                    alerts.append([float(misc["time"]), misc["name"]])
        ## Dedupe and order the additions by their time
        ## CBP should have these additions in reverse
        alerts = sorted(alerts, key=lambda x: x[0], reverse=True)
        hop_alerts = [
            [None, None],
            [None, None],
            [None, None],
            [None, None],
            [None, None],
            [None, None],
        ]
        for i in range(0, 6):
            try:
                if float(alerts[i][0]) > -1:
                    hop_alerts[i] = alerts[i]
            except:
                pass
        whirlpool_temps = sorted(whirlpool, key=lambda x: x[0], reverse=True)
        logging.info("Whirlpool Hops: {}".format(whirlpool_temps))
        try:
            whirlpool_temp = whirlpool_temps[0][0]
        except:
            whirlpool_temp = []
        logging.info("Whirlpool Temp: {}".format(whirlpool_temp))
        return hop_alerts, whirlpool_temp

    def getFirstWort(self, hops, recipe_type):
        alert = "No"
        names = []
        if recipe_type == "kbh":
            if len(hops) != 0:
                alert = "Yes"
                for hop in hops:
                    names.append(hop[1])
        elif recipe_type == "xml":
            for hop in hops:
                use = hop.find("USE").text
                ## Hops which are not used in the boil step should not cause alerts
                if use != "First Wort":
                    continue
                alert = "Yes"
                names.append(hop.find("NAME").text)
        elif recipe_type == "bf":
            for hop in hops:
                if hop["use"] == "First Wort":
                    alert = "Yes"
                    names.append(hop["name"])  ## TODO: Testing
        elif recipe_type == "json":
            if len(hops) != 0:
                alert = "Yes"
                for hop in hops:
                    names.append(hop["name"])

        return [alert, " and ".join(names)]

    async def create_Whirlpool_Cooldown(self, time: str = "15"):
        # Add Waitstep as Whirlpool
        if self.cooldown != "WaiStep" and self.cooldown != "":
            step_string = {
                "name": "Whirlpool",
                "props": {"Kettle": self.boilid, "Timer": time},
                "status_text": "",
                "status": "I",
                "type": "WaitStep",
            }
            await self.create_step(step_string)

        # CoolDown step is sending a notification when cooldowntemp is reached
        step_type = self.cooldown if self.cooldown != "" else "WaitStep"
        step_name = "CoolDown"
        cooldown_sensor = ""
        step_temp = ""
        step_timer = time
        if step_type.find("Cooldown") != -1:
            cooldown_sensor = self.cbpi.config.get("steps_cooldown_sensor", None)
            if cooldown_sensor is None or cooldown_sensor == "":
                cooldown_sensor = (
                    self.boilkettle.sensor
                )  # fall back to boilkettle sensor if no other sensor is specified
            step_timer = ""
            try:
                step_temp = (
                    int(self.CoolDownTemp)
                    if (
                        self.fermentation_step_temp is None
                        or self.fermentation_step_temp <= int(self.CoolDownTemp)
                    )
                    else self.fermentation_step_temp
                )
            except:
                step_temp = int(self.CoolDownTemp)
            step_string = {
                "name": "Cooldown",
                "props": {
                    "Kettle": self.boilid,
                    "Timer": step_timer,
                    "Temp": step_temp,
                    "Sensor": cooldown_sensor,
                    "Actor": self.CoolDownActor,
                },
                "status_text": "",
                "status": "I",
                "type": step_type,
            }
            await self.create_step(step_string)
      
    def get_config_values(self):
        self.kettle = None
        self.boilkettle = None
        # Define MashSteps
        self.TEMP_UNIT = self.cbpi.config.get("TEMP_UNIT", "C")
        self.AutoMode = self.cbpi.config.get("AutoMode", "Yes")
        self.mashin = self.cbpi.config.get("steps_mashin", "MashInStep")
        self.mash = self.cbpi.config.get("steps_mash", "MashStep")
        self.mashout = self.cbpi.config.get(
            "steps_mashout", None
        )  # Currently used only for the Braumeister
        self.boil = self.cbpi.config.get("steps_boil", "BoilStep")
        self.whirlpool = "Waitstep"
        self.cooldown = self.cbpi.config.get("steps_cooldown", "WaitStep")
        # get default boil temp from settings
        self.BoilTemp = self.cbpi.config.get("steps_boil_temp", 98)
        # get default cooldown temp alarm setting
        self.CoolDownTemp = self.cbpi.config.get("steps_cooldown_temp", 25)
        self.CoolDownActor = self.cbpi.config.get("steps_cooldown_actor", None)
        # get default Kettle from Settings
        self.id = self.cbpi.config.get("MASH_TUN", None)
        self.boilid = self.cbpi.config.get("BoilKettle", None)
        if self.boilid is None:
            self.boilid = self.id
        # If next parameter is Yes, MashIn Ste will be added before first mash step if not included in recipe
        self.addmashin = self.cbpi.config.get("AddMashInStep", "Yes")

        try:
            self.kettle = self.cbpi.kettle.find_by_id(self.id)
        except:
            self.cbpi.notify(
                "Recipe Upload",
                "No default Kettle defined. Please specify default Kettle in settings",
                NotificationType.ERROR,
            )
        try:
            self.boilkettle = self.cbpi.kettle.find_by_id(self.boilid)
        except:
            self.boilkettle = self.kettle

        config_values = {
            "kettle": self.kettle,
            "kettle_id": str(self.id),
            "boilkettle": self.boilkettle,
            "boilkettle_id": str(self.boilid),
            "mashin": str(self.mashin),
            "mash": str(self.mash),
            "mashout": str(self.mashout),
            "boil": str(self.boil),
            "whirlpool": str(self.whirlpool),
            "cooldown": str(self.cooldown),
            "boiltemp": str(self.BoilTemp),
            "cooldowntemp": str(self.CoolDownTemp),
            "cooldownactor": self.CoolDownActor,
            "temp_unit": str(self.TEMP_UNIT),
            "AutoMode": str(self.AutoMode),
        }
        logging.info(config_values)
        return config_values

    async def create_recipe(self, name):
        # Create recipe in recipe Book with name of first recipe in xml file
        self.recipeID = await self.cbpi.recipe.create(name)
        # send recipe to mash profile
        await self.cbpi.recipe.brew(self.recipeID)
        # remove empty recipe from recipe book
        await self.cbpi.recipe.remove(self.recipeID)

    # function to create json to be send to api to add a step to the current mash profile. Currently all properties are send to each step which does not cuase an issue
    async def create_step(self, step_string):
        # get server port from settings and define url for api calls -> adding steps
        self.port = str(self.cbpi.static_config.get("port", 8000))
        self.url = "http://127.0.0.1:" + self.port + "/step2/"
        # convert step:string to json required for api call.
        step = json.dumps(step_string)
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post(self.url, data=step) as response:
                return await response.text()
            await self.push_update()


__all__ = [
    "CBPiException",
    "KettleException",
    "FermenterException",
    "SensorException",
    "ActorException",
]


class CBPiException(Exception):
    pass


class KettleException(CBPiException):
    pass


class FermenterException(CBPiException):
    pass


class SensorException(CBPiException):
    pass


class ActorException(CBPiException):
    pass



# -*- coding: utf-8 -*-
import asyncio
import logging
from time import gmtime, strftime

from aiohttp import web
from cbpi.api import *
from cbpi.api import base
from cbpi.api.dataclasses import DataType, NotificationAction, NotificationType
from cbpi.api.timer import Timer

logger = logging.getLogger(__name__)


@parameters([])
class AlarmTimer(CBPiSensor):

    def __init__(self, cbpi, id, props):
        super(AlarmTimer, self).__init__(cbpi, id, props)
        self.value = "00:00:00"
        self.datatype = DataType.STRING
        self.timer = None
        self.time = 0
        self.stopped = False
        self.sensor = self.get_sensor(self.id)

    @action(
        key="Set Timer",
        parameters=[
            Property.Number(
                label="time", description="Time in Minutes", configurable=True
            )
        ],
    )
    async def set(self, time=0, **kwargs):
        self.stopped = False
        self.time = float(time)
        self.value = self.calculate_time(self.time)
        if self.timer is not None:
            await self.timer.stop()
        self.timer = Timer(
            int(self.time * 60),
            on_update=self.on_timer_update,
            on_done=self.on_timer_done,
        )
        await self.timer.stop()
        self.timer.is_running = False
        logging.info("Set Timer")

    @action(key="Start Timer", parameters=[])
    async def start(self, **kwargs):
        if self.timer is None:
            self.timer = Timer(
                int(self.time * 60),
                on_update=self.on_timer_update,
                on_done=self.on_timer_done,
            )

        if self.timer.is_running is not True:
            self.timer.start()
            self.stopped = False
            self.timer.is_running = True
        else:
            self.cbpi.notify(
                self.sensor.name, "Timer is already running", NotificationType.WARNING
            )

    @action(key="Stop Timer", parameters=[])
    async def stop(self, **kwargs):
        self.stopped = False
        await self.timer.stop()
        self.timer.is_running = False
        logging.info("Stop Timer")

    @action(key="Reset Timer", parameters=[])
    async def Reset(self, **kwargs):
        self.stopped = False
        if self.timer is not None:
            await self.timer.stop()
        self.value = self.calculate_time(self.time)
        self.timer = Timer(
            int(self.time * 60),
            on_update=self.on_timer_update,
            on_done=self.on_timer_done,
        )
        await self.timer.stop()
        self.timer.is_running = False
        logging.info("Reset Timer")

    async def on_timer_done(self, timer):
        # self.value = "Stopped"
        if self.stopped is True:
            self.cbpi.notify(self.sensor.name, "Timer done", NotificationType.SUCCESS)

        self.timer.is_running = False
        pass

    async def on_timer_update(self, timer, seconds):
        self.stopped = True
        self.value = Timer.format_time(seconds)

    async def run(self):
        while self.running is True:
            self.push_update(self.value)
            await asyncio.sleep(1)
            pass

    def get_state(self):
        return dict(value=self.value)

    def calculate_time(self, time):
        return strftime("%H:%M:%S", gmtime(time * 60))


def setup(cbpi):
    cbpi.plugin.register("AlarmTimer", AlarmTimer)
    pass



# -*- coding: utf-8 -*-
import asyncio
import logging
import re
import time
from datetime import datetime

from aiohttp import web
from cbpi.api import *
from cbpi.api.dataclasses import NotificationAction, NotificationType

cache = {}


@parameters(
    [
        Property.Text(label="Key", configurable=True, description="Http Key"),
        Property.Number(
            label="Timeout",
            configurable="True",
            unit="sec",
            description="Timeout in seconds to send notification (default:60 | deactivated: 0)",
        ),
        Property.Kettle(
            label="Kettle",
            description="Reduced logging if Kettle is inactive / range warning in dashboard (only Kettle or Fermenter to be selected)",
        ),
        Property.Fermenter(
            label="Fermenter",
            description="Reduced logging in seconds if Fermenter is inactive / range warning in dashboard (only Kettle or Fermenter to be selected)",
        ),
        Property.Number(
            label="ReducedLogging",
            configurable=True,
            description="Reduced logging frequency in seconds if selected Kettle or Fermenter is inactive (default: 60 sec | disabled: 0)",
        ),
        Property.Number(
            label="TempRange",
            configurable=True,
            unit="degree",
            description="Temp range in degree between reading and target temp of fermenter/kettle. Larger difference shows different color in dashboard (default:0 | deactivated: 0)",
        ),
    ]
)
class HTTPSensor(CBPiSensor):
    def __init__(self, cbpi, id, props):
        super(HTTPSensor, self).__init__(cbpi, id, props)
        self.running = True
        self.value = 0
        self.timeout = int(self.props.get("Timeout", 60))
        self.starttime = time.time()
        self.notificationsend = False
        self.nextchecktime = self.starttime + self.timeout
        self.sensor = self.get_sensor(self.id)
        self.lastdata = time.time()

        self.lastlog = 0
        self.reducedfrequency = int(self.props.get("ReducedLogging", 60))
        if self.reducedfrequency < 0:
            self.reducedfrequency = 0

        self.kettleid = self.props.get("Kettle", None)
        self.fermenterid = self.props.get("Fermenter", None)
        self.temprange = float(self.props.get("TempRange", 0))
        self.reducedlogging = True if self.kettleid or self.fermenterid else False

        if self.kettleid is not None and self.fermenterid is not None:
            self.reducedlogging = False
            self.cbpi.notify(
                "HTTPSensor",
                "Sensor '"
                + str(self.sensor.name)
                + "' cant't have Fermenter and Kettle defined for reduced logging / range warning",
                NotificationType.WARNING,
                action=[NotificationAction("OK", self.Confirm)],
            )

    async def Confirm(self, **kwargs):
        self.nextchecktime = time.time() + self.timeout
        self.notificationsend = False
        pass

    async def message(self):
        target_timestring = datetime.fromtimestamp(self.lastdata)
        self.cbpi.notify(
            "HTTPSensor Timeout",
            "Sensor '"
            + str(self.sensor.name)
            + "' did not respond.  Last data received: "
            + target_timestring.strftime("%D %H:%M"),
            NotificationType.WARNING,
            action=[NotificationAction("OK", self.Confirm)],
        )
        pass

    async def run(self):
        """
        This method is executed asynchronousely
        In this example the code is executed every second
        """
        while self.running is True:
            self.kettle = (
                self.get_kettle(self.kettleid) if self.kettleid is not None else None
            )
            self.fermenter = (
                self.get_fermenter(self.fermenterid)
                if self.fermenterid is not None
                else None
            )
            if self.timeout != 0:
                currenttime = time.time()
                if currenttime > self.nextchecktime and self.notificationsend == False:
                    await self.message()
                    self.notificationsend = True
            try:
                cache_value = cache.pop(self.props.get("Key"), None)
                if cache_value is not None:
                    self.value = float(cache_value)
                    self.push_update(self.value)

                    if self.reducedlogging:
                        await self.logvalue()
                    else:
                        self.log_data(self.value)
                        self.lastlog = time.time()

                    if self.timeout != 0:
                        self.nextchecktime = currenttime + self.timeout
                        self.notificationsend = False
                        self.lastdata = time.time()
            except Exception as e:
                logging.error(e)
                pass
            await asyncio.sleep(1)

    async def logvalue(self):
        now = time.time()
        if self.kettle is not None:
            try:
                kettlestatus = self.kettle.instance.state
            except:
                kettlestatus = False
            if kettlestatus:
                self.log_data(self.value)
                logging.info("Kettle Active")
                self.lastlog = time.time()
            else:
                logging.info("Kettle Inactive")
                if self.reducedfrequency != 0:
                    if now >= self.lastlog + self.reducedfrequency:
                        self.log_data(self.value)
                        self.lastlog = time.time()
                        logging.info("Logged with reduced freqency")
                        pass

        if self.fermenter is not None:
            try:
                fermenterstatus = self.fermenter.instance.state
            except:
                fermenterstatus = False
            if fermenterstatus:
                self.log_data(self.value)
                logging.info("Fermenter Active")
                self.lastlog = time.time()
            else:
                logging.info("Fermenter Inactive")
                if self.reducedfrequency != 0:
                    if now >= self.lastlog + self.reducedfrequency:
                        self.log_data(self.value)
                        self.lastlog = time.time()
                        logging.info("Logged with reduced freqency")
                        pass

    def get_state(self):
        # return the current state of the sensor
        return dict(value=self.value)


class HTTPSensorEndpoint(CBPiExtension):

    def __init__(self, cbpi):
        """
        Initializer

        :param cbpi:
        """
        self.pattern_check = re.compile("^[a-zA-Z0-9,.]{0,10}$")

        self.cbpi = cbpi
        # register component for http, events
        # In addtion the sub folder static is exposed to access static content via http
        self.cbpi.register(self, "/httpsensor")

    @request_mapping(path="/{key}/{value}", auth_required=False)
    async def http_new_value2(self, request):
        """
        ---
        description: Kettle Heater on
        tags:
        - HttpSensor
        parameters:
        - name: "key"
          in: "path"
          description: "Sensor Key"
          required: true
          type: "string"
        - name: "value"
          in: "path"
          description: "Value"
          required: true
          type: "integer"
          format: "int64"
        responses:
            "204":
                description: successful operation
        """

        global cache
        key = request.match_info["key"]
        value = request.match_info["value"]
        if self.pattern_check.match(key) is None:
            return web.json_response(
                status=422,
                data={"error": "Key not matching pattern ^[a-zA-Z0-9,.]{0,10}$"},
            )

        if self.pattern_check.match(value) is None:
            return web.json_response(
                status=422,
                data={"error": "Data not matching pattern ^[a-zA-Z0-9,.]{0,10}$"},
            )

        cache[key] = value

        return web.Response(status=204)


def setup(cbpi):
    """
    This method is called by the server during startup
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core
    :return:
    """
    cbpi.plugin.register("HTTPSensor", HTTPSensor)
    cbpi.plugin.register("HTTPSensorEndpoint", HTTPSensorEndpoint)



import logging
import os
import sys
from pathlib import Path
from zipfile import ZipFile

import requests
from cbpi import __codename__, __version__
from cbpi.configFolder import ConfigFolder
from cbpi.craftbeerpi import CraftBeerPi
from cbpi.utils.utils import load_config

try:
    import pwd

    module_pwd = True
except:
    module_pwd = False
import importlib
import pathlib
import pkgutil
import platform
import shutil
import time
from subprocess import call

import click
import inquirer
from colorama import Back, Fore, Style
from importlib_metadata import metadata
from tabulate import tabulate


class CraftBeerPiCli:
    def __init__(self, config) -> None:
        self.config = config
        pass

    def setup(self):
        print("Setting up CraftBeerPi")
        self.config.create_home_folder_structure()
        self.config.create_config_file()

    def start(self):
        if self.config.check_for_setup() is False:
            return
        print("START")
        cbpi = CraftBeerPi(self.config)
        cbpi.start()

    def setup_one_wire(self):
        print("Setting up 1Wire")
        with open("/boot/config.txt", "r") as f:
            lines = f.readlines()
        lines.append("dtoverlay=w1-gpio,gpiopin=4,pullup=on")

        configtempfile = os.path.join(self.config.get_file_path(""), "config.txt")

        with open(configtempfile, "w") as f:
            for line in lines:
                f.write(line)
        destfile = "/boot/config.txt"

        # copy and remove afterwards as mv will work, but raise an error message due to different file owners
        shutil.os.system('sudo cp "{}" "{}"'.format(configtempfile, destfile))
        shutil.os.system('rm -rf "{}"'.format(configtempfile))

        print("/boot/config.txt created")

    def list_one_wire(self):
        print("List 1Wire")
        call(["sudo", "modprobe", "w1-gpio"])
        call(["sudo", "modprobe", "w1-therm"])
        try:
            for dirname in os.listdir("/sys/bus/w1/devices"):
                if dirname.startswith("28") or dirname.startswith("10"):
                    print(dirname)
        except Exception as e:
            print(e)

    def plugins_list(self):
        result = []
        print("")
        print(Fore.LIGHTYELLOW_EX, "List of active plugins", Style.RESET_ALL)
        print("")
        discovered_plugins = {
            name: importlib.import_module(name)
            for finder, name, ispkg in pkgutil.iter_modules()
            if name.startswith("cbpi") and len(name) > 4
        }
        for key, module in discovered_plugins.items():
            try:
                meta = metadata(key)
                result.append(
                    dict(
                        Name=meta["Name"],
                        Version=meta["Version"],
                        Author=meta["Author"],
                        Homepage=meta["Home-page"],
                        Summary=meta["Summary"],
                    )
                )

            except Exception as e:
                print(e)
        print(Fore.LIGHTGREEN_EX, tabulate(result, headers="keys"), Style.RESET_ALL)

    def plugin_create(self, pluginName):
        print("Plugin Creation")
        print("")

        questions = [
            inquirer.Text(
                "name",
                message='Plugin Name (will be prefixed with "cbpi4-")',
                default=pluginName,
            ),
        ]

        answers = inquirer.prompt(questions)

        if answers["name"] == "":
            print("you failed to provide a name for the new plugin - terminating")
            return

        name = "cbpi4-" + str(answers["name"]).replace("_", "-").replace(" ", "-")
        if os.path.exists(os.path.join(".", name)) is True:
            print("Cant create Plugin. Folder {} already exists ".format(name))
            return

        url = (
            "https://github.com/PiBrewing/craftbeerpi4-plugin-template/archive/main.zip"
        )
        r = requests.get(url)
        with open("temp.zip", "wb") as f:
            f.write(r.content)

        with ZipFile("temp.zip", "r") as repo_zip:
            repo_zip.extractall()

        time.sleep(4)  # windows dev container permissions problem otherwise

        os.rename(
            os.path.join(".", "craftbeerpi4-plugin-template-main"),
            os.path.join(".", name),
        )
        os.rename(os.path.join(".", name, "src"), os.path.join(".", name, name))

        import jinja2

        templateLoader = jinja2.FileSystemLoader(searchpath=os.path.join(".", name))
        templateEnv = jinja2.Environment(loader=templateLoader)
        TEMPLATE_FILE = "setup.py"
        template = templateEnv.get_template(TEMPLATE_FILE)
        outputText = template.render(name=name)

        with open(os.path.join(".", name, "setup.py"), "w") as fh:
            fh.write(outputText)

        TEMPLATE_FILE = "MANIFEST.in"
        template = templateEnv.get_template(TEMPLATE_FILE)
        outputText = template.render(name=name)
        with open(os.path.join(".", name, "MANIFEST.in"), "w") as fh:
            fh.write(outputText)

        TEMPLATE_FILE = os.path.join("/", name, "config.yaml")
        operatingsystem = str(platform.system()).lower()
        if operatingsystem.startswith("win"):
            TEMPLATE_FILE = str(TEMPLATE_FILE).replace("\\", "/")

        template = templateEnv.get_template(TEMPLATE_FILE)
        outputText = template.render(name=name)

        with open(os.path.join(".", name, name, "config.yaml"), "w") as fh:
            fh.write(outputText)

        print("")
        print("")
        print(
            "Plugin {}{}{} created! ".format(Fore.LIGHTGREEN_EX, name, Style.RESET_ALL)
        )
        print("")
        print(
            "Developer Documentation: https://openbrewing.gitbook.io/craftbeerpi4_support/readme/development"
        )
        print("")
        print("Happy developing! Cheers")
        print("")
        print("")

    def autostart(self, name):
        """Enable or disable autostart"""
        if name == "status":
            if (
                os.path.exists(
                    os.path.join("/etc/systemd/system", "craftbeerpi.service")
                )
                is True
            ):
                print(
                    "CraftBeerPi Autostart is {}ON{}".format(
                        Fore.LIGHTGREEN_EX, Style.RESET_ALL
                    )
                )
            else:
                print(
                    "CraftBeerPi Autostart is {}OFF{}".format(Fore.RED, Style.RESET_ALL)
                )
        elif name == "on":
            # user=os.getlogin()
            user = pwd.getpwuid(os.getuid()).pw_name
            path = "/usr/local/bin/cbpi"
            if os.path.exists("/home/" + user + "/.local/bin/cbpi") is True:
                path = "/home/" + user + "/.local/bin/cbpi"
            print("Add craftbeerpi.service to systemd")
            try:
                if (
                    os.path.exists(
                        os.path.join("/etc/systemd/system", "craftbeerpi.service")
                    )
                    is False
                ):
                    templatefile = self.config.get_file_path("craftbeerpi.template")
                    shutil.os.system(
                        'cp "{}" "{}"'.format(
                            templatefile,
                            self.config.get_file_path("craftbeerpi.service"),
                        )
                    )
                    srcfile = self.config.get_file_path("craftbeerpi.service")
                    import jinja2

                    templateLoader = jinja2.FileSystemLoader(
                        searchpath=os.path.join(self.config.get_file_path(""))
                    )
                    templateEnv = jinja2.Environment(loader=templateLoader)
                    operatingsystem = str(platform.system()).lower()
                    if operatingsystem.startswith("win"):
                        srcfile = str(srcfile).replace("\\", "/")

                    template = templateEnv.get_template("craftbeerpi.service")
                    outputText = template.render(user=user, path=path)
                    with open(srcfile, "w") as fh:
                        fh.write(outputText)
                    destfile = os.path.join("/etc/systemd/system")
                    shutil.os.system('sudo mv "{}" "{}"'.format(srcfile, destfile))
                    print("Copied craftbeerpi.service to /etc/systemd/system")
                    shutil.os.system("sudo systemctl enable craftbeerpi.service")
                    print("Enabled craftbeerpi service")
                    shutil.os.system("sudo systemctl start craftbeerpi.service")
                    print("Started craftbeerpi.service")
                else:
                    print(
                        "craftbeerpi.service is already located in /etc/systemd/system"
                    )
            except Exception as e:
                print(e)
                return
            return
        elif name == "off":
            print("Remove craftbeerpi.service from systemd")
            try:
                status = os.popen(
                    "systemctl list-units --type=service --state=running | grep craftbeerpi.service"
                ).read()
                if status.find("craftbeerpi.service") != -1:
                    shutil.os.system("sudo systemctl stop craftbeerpi.service")
                    print("Stopped craftbeerpi service")
                    shutil.os.system("sudo systemctl disable craftbeerpi.service")
                    print("Removed craftbeerpi.service as service")
                else:
                    print("craftbeerpi.service service is not running")

                if (
                    os.path.exists(
                        os.path.join("/etc/systemd/system", "craftbeerpi.service")
                    )
                    is True
                ):
                    shutil.os.system(
                        'sudo rm -rf "{}"'.format(
                            os.path.join("/etc/systemd/system", "craftbeerpi.service")
                        )
                    )
                    print("Deleted craftbeerpi.service from /etc/systemd/system")
                else:
                    print("craftbeerpi.service is not located in /etc/systemd/system")
            except Exception as e:
                print(e)
                return
            return

    def chromium(self, name):
        """Enable or disable autostart"""
        if name == "status":
            if (
                os.path.exists(os.path.join("/etc/xdg/autostart/", "chromium.desktop"))
                is True
            ):
                print(
                    "CraftBeerPi Chromium Desktop is {}ON{}".format(
                        Fore.LIGHTGREEN_EX, Style.RESET_ALL
                    )
                )
            else:
                print(
                    "CraftBeerPi Chromium Desktop is {}OFF{}".format(
                        Fore.RED, Style.RESET_ALL
                    )
                )
        elif name == "on":
            print("Add chromium.desktop to /etc/xdg/autostart/")
            try:
                if (
                    os.path.exists(
                        os.path.join("/etc/xdg/autostart/", "chromium.desktop")
                    )
                    is False
                ):
                    srcfile = self.config.get_file_path("chromium.desktop")
                    destfile = os.path.join("/etc/xdg/autostart/")
                    shutil.os.system('sudo cp "{}" "{}"'.format(srcfile, destfile))
                    print("Copied chromium.desktop to /etc/xdg/autostart/")
                else:
                    print("chromium.desktop is already located in /etc/xdg/autostart/")
            except Exception as e:
                print(e)
                return
            return
        elif name == "off":
            print("Remove chromium.desktop from /etc/xdg/autostart/")
            try:
                if (
                    os.path.exists(
                        os.path.join("/etc/xdg/autostart/", "chromium.desktop")
                    )
                    is True
                ):
                    shutil.os.system(
                        'sudo rm -rf "{}"'.format(
                            os.path.join("/etc/xdg/autostart/", "chromium.desktop")
                        )
                    )
                    print("Deleted chromium.desktop from /etc/xdg/autostart/")
                else:
                    print("chromium.desktop is not located in /etc/xdg/autostart/")
            except Exception as e:
                print(e)
                return
            return


@click.group()
@click.pass_context
@click.option(
    "--config-folder-path",
    "-c",
    default="./config",
    type=click.Path(),
    help="Specify where the config folder is located. Defaults to './config'.",
)
@click.option(
    "--logs-folder-path",
    "-l",
    default="",
    type=click.Path(),
    help="Specify where the log folder is located. Defaults to '../logs' relative from the config folder.",
)
@click.option(
    "--debug-log-level",
    "-d",
    default="99",
    type=int,
    help="Specify the log level you want to write to all logs. 0=ALL, 10=DEBUG, 20=INFO 30(default)=WARNING, 40=ERROR, 50=CRITICAL. Can be also set in config.yaml (debug-log-level: INT)",
)
def main(context, config_folder_path, logs_folder_path, debug_log_level):
    print("--------------------------")
    print("Welcome to CBPi " + __version__)
    print("--------------------------")
    if logs_folder_path == "":
        logs_folder_path = os.path.join(
            Path(config_folder_path).absolute().parent, "logs"
        )
    formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
    )
    config = ConfigFolder(config_folder_path, logs_folder_path)
    static_config = load_config(config.get_file_path("config.yaml"))
    try:
        if debug_log_level == 99:
            debug_log_level = static_config["debug-log-level"]
    except:
        debug_log_level = 30

    logging.basicConfig(format=formatter, stream=logging.StreamHandler())
    logger = logging.getLogger()
    print("*******************************")
    print("Debug-log-level is {}".format(debug_log_level))
    print("*******************************")
    logger.setLevel(debug_log_level)
    try:
        if not os.path.isdir(logs_folder_path):
            logger.info(
                f"logs folder '{logs_folder_path}' doesnt exist and we are trying to create it"
            )
            pathlib.Path(logs_folder_path).mkdir(parents=True, exist_ok=True)
            logger.info(f"logs folder '{logs_folder_path}' successfully created")
        handler = logging.handlers.RotatingFileHandler(
            os.path.join(logs_folder_path, f"cbpi.log"), maxBytes=1000000, backupCount=3
        )
        logger.addHandler(handler)
        handler.setFormatter(formatter)
    except Exception as e:
        logger.warning(
            "log folder or log file could not be created or accessed. check folder and file permissions or create the logs folder somewhere you have access with a start option like '--log-folder-path=./logs'"
        )
        logging.critical(e, exc_info=True)
    cbpi_cli = CraftBeerPiCli(config)
    context.obj = cbpi_cli


@main.command()
@click.pass_context
def setup(context):
    """Create Config folder"""
    context.obj.setup()


@main.command()
@click.pass_context
@click.option("--list", is_flag=True, help="List all 1Wire Devices")
@click.option("--setup", is_flag=True, help="Setup 1Wire on Raspberry Pi")
def onewire(context, list, setup):
    """(--setup | --list) Setup 1wire on Raspberry Pi or list sensors"""
    operationsystem = sys.platform
    if not operationsystem.startswith("win"):
        if setup is True:
            context.obj.setup_one_wire()
        if list is True:
            context.obj.list_one_wire()
    else:
        print("Onewire options NOT available under Windows")


@main.command()
@click.pass_context
def start(context):
    context.obj.start()


@main.command()
@click.pass_context
def plugins(context):
    """List active plugins"""
    context.obj.plugins_list()


@main.command()
@click.pass_context
@click.argument("pluginname", nargs=-1, required=False)
def create(context, pluginname=[]):
    """Create New Plugin"""
    sentence = ""
    for word in pluginname:
        if sentence != "":
            sentence += " "
        sentence += word
    context.obj.plugin_create(sentence)


@main.command()
@click.pass_context
@click.argument("name")
def autostart(context, name):
    """(on|off|status) Enable or disable autostart"""
    operationsystem = sys.platform
    if not operationsystem.startswith("win"):
        context.obj.autostart(name)
    else:
        print("Autostart option NOT available under Windows")


@main.command()
@click.pass_context
@click.argument("name")
def chromium(context, name):
    """(on|off|status) Enable or disable Kiosk mode"""
    operationsystem = sys.platform
    if not operationsystem.startswith("win"):
        context.obj.chromium(name)
    else:
        print("Chromium option NOT available under Windows")



import glob
import importlib
import json
import logging
import os
import pathlib
import pkgutil
import shutil
import socket
import zipfile
from datetime import date, datetime, timedelta

import aiohttp
import psutil
from cbpi.api import *
from cbpi.api.base import CBPiBase
from cbpi.api.config import ConfigType
from cbpi.api.dataclasses import NotificationAction, NotificationType
from tabulate import tabulate
from voluptuous.schema_builder import message

try:
    from systemd import journal

    systemd_available = True
except Exception:
    logging.warning("Failed to load systemd library. logfile download not available")
    systemd_available = False


class SystemController:

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self.service = cbpi.actor
        self.logger = logging.getLogger(__name__)
        self.logsFolderPath = self.cbpi.config_folder.logsFolderPath

        self.cbpi.app.on_startup.append(self.check_for_update)

    async def check_for_update(self, app):
        pass

    async def restart(self):
        logging.info("RESTART")
        os.system("systemctl reboot")
        pass

    async def shutdown(self):
        logging.info("SHUTDOWN")
        os.system("systemctl poweroff")
        pass

    async def backupConfig(self):
        files = glob.glob("*cbpi4_config*.zip")
        for f in files:
            try:
                os.remove(f)
            except Exception as e:
                logging.error("Cannot remove old config backup: {}".format(e))

        try:
            current_date = date.today()
            current_date = str(current_date).replace("-", "_")
            output_filename = current_date + "_cbpi4_config"
        except:
            output_filename = "cbpi4_config"
        dir_name = pathlib.Path(self.cbpi.config_folder.get_file_path(""))
        shutil.make_archive(output_filename, "zip", dir_name)
        return output_filename + ".zip"

    async def plugins_list(self):
        result = []
        discovered_plugins = {
            name: importlib.import_module(name)
            for finder, name, ispkg in pkgutil.iter_modules()
            if name.startswith("cbpi") and len(name) > 4
        }
        for key, module in discovered_plugins.items():
            from importlib.metadata import version

            try:
                from importlib.metadata import distribution, metadata, version

                meta = metadata(key)
                result.append(
                    dict(
                        Name=meta["Name"],
                        Version=meta["Version"],
                        Author=meta["Author"],
                        Homepage=meta["Home-page"],
                        Summary=meta["Summary"],
                    )
                )

            except Exception as e:
                print(e)
        return tabulate(result, headers="keys")

    async def downloadlog(self, logtime):
        filename = "cbpi4.log"
        fullname = pathlib.Path(os.path.join(".", filename))
        pluginname = "cbpi4_plugins.txt"
        fullpluginname = pathlib.Path(os.path.join(".", pluginname))
        actorname = "cbpi4_actors.txt"
        fullactorname = pathlib.Path(os.path.join(".", actorname))
        sensorname = "cbpi4_sensors.txt"
        fullsensorname = pathlib.Path(os.path.join(".", sensorname))
        kettlename = "cbpi4_kettles.txt"
        fullkettlename = pathlib.Path(os.path.join(".", kettlename))

        output_filename = "cbpi4_log.zip"
        result = []
        if systemd_available:
            j = journal.Reader()
            if logtime == "b":
                j.this_boot()
            else:
                since = datetime.now() - timedelta(hours=int(logtime))
                j.seek_realtime(since)
            j.add_match(_SYSTEMD_UNIT="craftbeerpi.service")

            for entry in j:
                result.append(entry["MESSAGE"])
        else:
            try:
                logfilename = pathlib.Path(self.logsFolderPath + "/" + "cbpi.log")
                with open(logfilename) as f:
                    for line in f:
                        result.append(line.rstrip("\n"))
            except:
                pass

        try:
            with open(fullname, "w") as f:
                for line in result:
                    f.write(f"{line}\n")
        except Exception as e:
            logging.error(e)

        plugins = await self.plugins_list()
        with open(fullpluginname, "w") as f:
            f.write(plugins)

        try:
            actors = self.cbpi.actor.get_state()
            json.dump(
                actors["data"], open(fullactorname, "w"), indent=4, sort_keys=True
            )
            sensors = self.cbpi.sensor.get_state()
            json.dump(
                sensors["data"], open(fullsensorname, "w"), indent=4, sort_keys=True
            )
            kettles = self.cbpi.kettle.get_state()
            json.dump(
                kettles["data"], open(fullkettlename, "w"), indent=4, sort_keys=True
            )
        except Exception as e:
            logging.info(e)
            self.cbpi.notify(
                "Error",
                "Creation of files failed: {}".format(e),
                NotificationType.ERROR,
            )

        try:
            zipObj = zipfile.ZipFile(output_filename, "w", zipfile.ZIP_DEFLATED)
            zipObj.write(fullname)
            zipObj.write(fullpluginname)
            zipObj.write(fullactorname)
            zipObj.write(fullsensorname)
            zipObj.write(fullkettlename)
            zipObj.close()
        except Exception as e:
            logging.info(e)
            self.cbpi.notify(
                "Error", "Zip creation failed: {}".format(e), NotificationType.ERROR
            )

        try:
            os.remove(fullname)
            os.remove(fullpluginname)
            os.remove(fullactorname)
            os.remove(fullsensorname)
            os.remove(fullkettlename)
        except Exception as e:
            logging.info(e)
            self.cbpi.notify(
                "Error",
                "Removal of original files failed: {}".format(e),
                NotificationType.ERROR,
            )

    def allowed_file(self, filename, extension):
        return "." in filename and filename.rsplit(".", 1)[1] in set([extension])

    def recursive_chown(self, path, owner, group):
        for dirpath, dirnames, filenames in os.walk(path):
            shutil.chown(dirpath, owner, group)
            for filename in filenames:
                shutil.chown(os.path.join(dirpath, filename), owner, group)

    async def restoreConfig(self, data):
        fileData = data["File"]
        filename = fileData.filename
        backup_file = fileData.file
        content_type = fileData.content_type
        required_content = [
            "dashboard/",
            "recipes/",
            "upload/",
            "config.json",
            "config.yaml",
        ]

        if content_type == "application/x-zip-compressed" or content_type == "application/zip":
            try:
                content = backup_file.read()
                if backup_file and self.allowed_file(filename, "zip"):
                    self.path = os.path.join(
                        self.cbpi.config_folder.configFolderPath, "restored_config.zip"
                    )

                    f = open(self.path, "wb")
                    f.write(content)
                    f.close()
                    zip = zipfile.ZipFile(self.path)
                    zip_content_list = zip.namelist()
                    zip_content = True
                    for content in required_content:
                        try:
                            check = zip_content_list.index(content)
                        except:
                            zip_content = False
                    if zip_content == True:
                        self.cbpi.notify(
                            "Success",
                            "Config backup has been uploaded",
                            NotificationType.SUCCESS,
                        )
                        self.cbpi.notify(
                            "Action Required!",
                            "Please restart the server",
                            NotificationType.WARNING,
                        )
                    else:
                        self.cbpi.notify(
                            "Error",
                            "Wrong content type. Upload failed",
                            NotificationType.ERROR,
                        )
                        os.remove(self.path)
            except:
                self.cbpi.notify(
                    "Error", "Config backup upload failed", NotificationType.ERROR
                )
                pass
        else:
            self.cbpi.notify(
                "Error", "Wrong content type. Upload failed", NotificationType.ERROR
            )

    async def uploadSVG(self, data):
        fileData = data["File"]
        filename = fileData.filename
        svg_file = fileData.file
        content_type = fileData.content_type

        logging.info(content_type)

        if content_type == "image/svg+xml":
            try:
                content = svg_file.read().decode("utf-8", "replace")
                if svg_file and self.allowed_file(filename, "svg"):
                    self.path = os.path.join(
                        self.cbpi.config_folder.get_file_path("dashboard"),
                        "widgets",
                        filename,
                    )
                    logging.info(self.path)

                    f = open(self.path, "w")
                    f.write(content)
                    f.close()
                    self.cbpi.notify(
                        "Success",
                        "SVG file ({}) has been uploaded.".format(filename),
                        NotificationType.SUCCESS,
                    )
            except:
                self.cbpi.notify("Error", "SVG upload failed", NotificationType.ERROR)
                pass
        else:
            self.cbpi.notify(
                "Error", "Wrong content type. Upload failed", NotificationType.ERROR
            )

    async def systeminfo(self):
        logging.info("SYSTEMINFO")
        system = ""
        temp = 0
        cpuload = 0
        cpucount = 0
        cpufreq = 0
        totalmem = 0
        availmem = 0
        mempercent = 0
        eth0IP = "N/A"
        wlan0IP = "N/A"
        eth0speed = "N/A"
        wlan0speed = "N/A"

        TEMP_UNIT = self.cbpi.config.get("TEMP_UNIT", "C")
        FAHRENHEIT = False if TEMP_UNIT == "C" else True

        af_map = {
            socket.AF_INET: "IPv4",
            socket.AF_INET6: "IPv6",
        }

        try:
            if psutil.LINUX == True:
                system = "Linux"
            elif psutil.WINDOWS == True:
                system = "Windows"
            elif psutil.MACOS == True:
                system = "MacOS"
            cpuload = round(psutil.cpu_percent(interval=None), 1)
            cpucount = psutil.cpu_count(logical=False)
            cpufreq = psutil.cpu_freq()
            mem = psutil.virtual_memory()
            availmem = round((int(mem.available) / (1024 * 1024)), 1)
            mempercent = round(float(mem.percent), 1)
            totalmem = round((int(mem.total) / (1024 * 1024)), 1)
            if system == "Linux":
                try:
                    temps = psutil.sensors_temperatures(fahrenheit=FAHRENHEIT)
                    for name, entries in temps.items():
                        for entry in entries:
                            if name == "cpu_thermal":
                                temp = round(float(entry.current), 1)
                except:
                    pass
            else:
                temp = "N/A"
            if system == "Linux":
                try:
                    ethernet = psutil.net_if_addrs()
                    for nic, addrs in ethernet.items():
                        if nic == "eth0":
                            for addr in addrs:
                                if (
                                    str(addr.family) == "AddressFamily.AF_INET"
                                    or str(addr.family) == "2"
                                ):
                                    if addr.address:
                                        eth0IP = addr.address
                        if nic == "wlan0":
                            for addr in addrs:
                                if (
                                    str(addr.family) == "AddressFamily.AF_INET"
                                    or str(addr.family) == "2"
                                ):
                                    if addr.address:
                                        wlan0IP = addr.address
                    info = psutil.net_if_stats()
                    try:
                        for nic in info:
                            if nic == "eth0":
                                if info[nic].isup == True:
                                    if info[nic].speed:
                                        eth0speed = info[nic].speed
                                else:
                                    eth0speed = "down"
                            if nic == "wlan0":
                                if info[nic].isup == True:
                                    ratestring = os.popen(
                                        "iwlist wlan0 rate | grep Rate"
                                    ).read()
                                    start = ratestring.find("=") + 1
                                    end = ratestring.find(" Mb/s")
                                    wlan0speed = ratestring[start:end]
                                else:
                                    wlan0speed = "down"
                    except Exception as e:
                        logging.info(e)
                except:
                    pass

            if system == "Windows":
                try:
                    ethernet = psutil.net_if_addrs()
                    for nic, addrs in ethernet.items():
                        if nic == "Ethernet":
                            for addr in addrs:
                                if str(addr.family) == "AddressFamily.AF_INET":
                                    if addr.address:
                                        eth0IP = addr.address
                        if nic == "WLAN":
                            for addr in addrs:
                                if str(addr.family) == "AddressFamily.AF_INET":
                                    if addr.address:
                                        wlan0IP = addr.address
                    info = psutil.net_if_stats()
                    try:
                        for nic in info:
                            if nic == "Ethernet":
                                if info[nic].isup == True:
                                    if info[nic].speed:
                                        eth0speed = info[nic].speed
                                else:
                                    eth0speed = "down"
                            if nic == "WLAN":
                                if info[nic].isup == True:
                                    if info[nic].speed:
                                        wlan0speed = info[nic].speed
                                else:
                                    wlan0speed = "down"
                    except Exception as e:
                        logging.info(e)
                except:
                    pass

        except:
            pass

        systeminfo = {
            "system": system,
            "cpuload": cpuload,
            "cpucount": cpucount,
            "cpufreq": cpufreq.current,
            "totalmem": totalmem,
            "availmem": availmem,
            "mempercent": mempercent,
            "temp": temp,
            "temp_unit": TEMP_UNIT,
            "eth0": eth0IP,
            "wlan0": wlan0IP,
            "eth0speed": eth0speed,
            "wlan0speed": wlan0speed,
        }
        return systeminfo



import logging
from datetime import datetime
from socket import timeout
from typing import KeysView
from unittest.mock import MagicMock, patch

from cbpi.api import *
from cbpi.api.dataclasses import NotificationAction, NotificationType
from voluptuous.schema_builder import message

logger = logging.getLogger(__name__)


@parameters([])
class DummyActor(CBPiActor):

    def __init__(self, cbpi, id, props):
        super().__init__(cbpi, id, props)

    @action("SAY HELLO", {})
    async def helloWorld(self, **kwargs):
        self.cbpi.notify("HELLO", "WOOHO", NotificationType.ERROR)

    async def start(self):
        await super().start()

    async def on(self, power=0):
        logger.info("ACTOR %s ON " % self.id)
        self.state = True

    async def off(self):
        logger.info("ACTOR %s OFF " % self.id)

        self.state = False

    def get_state(self):

        return self.state

    async def run(self):
        pass


def setup(cbpi):
    """
    This method is called by the server during startup
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core
    :return:
    """

    cbpi.plugin.register("DummyActor", DummyActor)



# -*- coding: utf-8 -*-
import asyncio
import logging
import os
import threading
import time
from subprocess import call

from aiohttp import web
from cbpi.api import *
from cbpi.api.dataclasses import NotificationAction, NotificationType


def getSensors():
    try:
        arr = []
        for dirname in os.listdir("/sys/bus/w1/devices"):
            if dirname.startswith("28") or dirname.startswith("10"):
                arr.append(dirname)
        return arr
    except:
        return []


class ReadThread(threading.Thread):

    value = 0

    def __init__(self, sensor_name):
        threading.Thread.__init__(self)
        self.value = 0
        self.sensor_name = sensor_name
        self.runnig = True

    def shutdown(self):
        pass

    def stop(self):
        self.runnig = False

    def run(self):

        while self.runnig:
            try:
                if self.sensor_name is None:
                    return
                with open(
                    "/sys/bus/w1/devices/%s/w1_slave" % self.sensor_name, "r"
                ) as content_file:
                    content = content_file.read()
                    if content.split("\n")[0].split(" ")[11] == "YES":
                        temp = float(content.split("=")[-1]) / 1000  # temp in Celcius
                        self.value = temp
            except:
                pass

            time.sleep(1)


@parameters(
    [
        Property.Select(label="Sensor", options=getSensors()),
        Property.Number(
            label="offset",
            configurable=True,
            default_value=0,
            description="Sensor Offset (Default is 0)",
        ),
        Property.Select(
            label="Interval",
            options=[1, 5, 10, 30, 60],
            description="Interval in Seconds",
        ),
        Property.Kettle(
            label="Kettle",
            description="Reduced logging if Kettle is inactive / range warning in dashboard(only Kettle or Fermenter to be selected)",
        ),
        Property.Fermenter(
            label="Fermenter",
            description="Reduced logging in seconds if Fermenter is inactive / range warning in dashboard (only Kettle or Fermenter to be selected)",
        ),
        Property.Number(
            label="ReducedLogging",
            configurable=True,
            description="Reduced logging frequency in seconds if selected Kettle or Fermenter is inactive (default: 60 sec | disabled: 0)",
        ),
        Property.Number(
            label="TempRange",
            configurable=True,
            unit="degree",
            description="Temp range in degree between reading and target temp of fermenter/kettle. Larger difference shows different color in dashboard (default:0 | deactivated: 0)",
        ),
    ]
)
class OneWire(CBPiSensor):

    def __init__(self, cbpi, id, props):
        super(OneWire, self).__init__(cbpi, id, props)
        self.value = 200

    async def start(self):
        await super().start()
        self.name = self.props.get("Sensor")
        self.interval = int(self.props.get("Interval", 60))
        self.offset = float(self.props.get("offset", 0))

        self.reducedfrequency = float(self.props.get("ReducedLogging", 60))
        if self.reducedfrequency < 0:
            self.reducedfrequency = 0
        self.lastlog = 0
        self.temprange = float(self.props.get("TempRange", 0))
        self.sensor = self.get_sensor(self.id)
        self.kettleid = self.props.get("Kettle", None)
        self.fermenterid = self.props.get("Fermenter", None)
        self.reducedlogging = True if self.kettleid or self.fermenterid else False

        if self.kettleid is not None and self.fermenterid is not None:
            self.reducedlogging = False
            self.cbpi.notify(
                "OneWire Sensor",
                "Sensor '"
                + str(self.sensor.name)
                + "' cant't have Fermenter and Kettle defined for reduced logging / range warning.",
                NotificationType.WARNING,
                action=[NotificationAction("OK", self.Confirm)],
            )
        if (self.reducedfrequency != 0) and (self.interval >= self.reducedfrequency):
            self.reducedlogging = False
            self.cbpi.notify(
                "OneWire Sensor",
                "Sensor '"
                + str(self.sensor.name)
                + "' has shorter or equal 'reduced logging' compared to regular interval.",
                NotificationType.WARNING,
                action=[NotificationAction("OK", self.Confirm)],
            )

        self.t = ReadThread(self.name)
        self.t.daemon = True

        def shutdown():
            shutdown.cb.shutdown()

        shutdown.cb = self.t
        self.t.start()

    async def Confirm(self, **kwargs):
        pass

    async def stop(self):
        try:
            self.t.stop()
            self.running = False
        except:
            pass

    async def run(self):

        self.kettle = (
            self.get_kettle(self.kettleid) if self.kettleid is not None else None
        )
        self.fermenter = (
            self.get_fermenter(self.fermenterid)
            if self.fermenterid is not None
            else None
        )

        while self.running == True:
            self.TEMP_UNIT = self.get_config_value("TEMP_UNIT", "C")
            if (
                self.TEMP_UNIT == "C"
            ):  # Report temp in C if nothing else is selected in settings
                self.value = round((self.t.value + self.offset), 2)
            else:  # Report temp in F if unit selected in settings
                self.value = round((9.0 / 5.0 * self.t.value + 32 + self.offset), 2)
            self.push_update(self.value)
            if self.reducedlogging:
                await self.logvalue()
            else:
                logging.info("OneWire {} regular logging".format(self.sensor.name))
                self.log_data(self.value)
                self.lastlog = time.time()
            await asyncio.sleep(self.interval)

    async def logvalue(self):
        now = time.time()
        logging.info("OneWire {} logging subroutine".format(self.sensor.name))
        if self.kettle is not None:
            try:
                kettlestatus = self.kettle.instance.state
            except:
                kettlestatus = False
            if kettlestatus:
                self.log_data(self.value)
                logging.info("OneWire {} Kettle Active".format(self.sensor.name))
                self.lastlog = time.time()
            else:
                logging.info("OneWire {} Kettle Inactive".format(self.sensor.name))
                if self.reducedfrequency != 0:
                    if now >= self.lastlog + self.reducedfrequency:
                        self.log_data(self.value)
                        self.lastlog = time.time()
                        logging.info("Logged with reduced freqency")
                        pass

        if self.fermenter is not None:
            try:
                fermenterstatus = self.fermenter.instance.state
            except:
                fermenterstatus = False
            if fermenterstatus:
                self.log_data(self.value)
                logging.info("OneWire {} Fermenter Active".format(self.sensor.name))
                self.lastlog = time.time()
            else:
                logging.info("OneWire {} Fermenter Inactive".format(self.sensor.name))
                if self.reducedfrequency != 0:
                    if now >= self.lastlog + self.reducedfrequency:
                        self.log_data(self.value)
                        self.lastlog = time.time()
                        logging.info("Logged with reduced freqency")
                        pass

    def get_state(self):
        return dict(value=self.value)


def setup(cbpi):
    cbpi.plugin.register("OneWire", OneWire)
    try:
        # Global Init
        call(["sudo", "modprobe", "w1-gpio"])
        call(["sudo", "modprobe", "w1-therm"])
    except Exception as e:
        pass



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, List

from cbpi.api.config import ConfigType
from cbpi.api.step import StepState


class Props:

    def __init__(self, data={}):
        super(Props, self).__setattr__("__data__", {})
        for key, value in data.items():
            self.__setattr__(key, value)

    def __getattr__(self, name):
        return self.__data__.get(name)

    def __setattr__(self, name, value):
        self.__data__[name] = value

    def __str__(self):
        return self.__data__.__str__()

    def __getitem__(self, key):
        return self.__data__[key]

    def __setitem__(self, key, value):
        self.__data__[key] = value

    def __contains__(self, key):
        return key in self.__data__

    def get(self, key, d=None):
        if key in self.__data__ and self.__data__[key] != "":
            return self.__data__[key]
        else:
            return d

    def to_dict(self):
        def parse_object(value):
            if isinstance(value, Props):
                return value.to_dict()
            elif isinstance(value, list):
                return list(map(parse_object, value))
            else:
                return value

        return dict(
            (key, parse_object(value)) for (key, value) in self.__data__.items()
        )


@dataclass
class Actor:
    id: str = None
    name: str = None
    props: Props = Props()
    state: bool = False
    power: int = 100
    timer: int = 0
    type: str = None
    instance: str = None

    def __str__(self):
        return "name={} props={}, state={}, type={}, power={}, timer={}".format(
            self.name, self.props, self.state, self.type, self.power, self.timer
        )

    def to_dict(self):
        if self.instance is not None:
            state = self.instance.get_state()
            actortype = self.type
        else:
            state = False
            actortype = self.type  # !!! MISSING TYPE !!!
        return dict(
            id=self.id,
            name=self.name,
            type=actortype,
            props=self.props.to_dict(),
            state=state,
            power=self.power,
            timer=self.timer,
        )


class DataType(Enum):
    VALUE = "value"
    DATETIME = "datetime"
    STRING = "string"


@dataclass
class Sensor:
    id: str = None
    name: str = None
    props: Props = Props()
    state: bool = False
    type: str = None
    instance: str = None
    datatype: DataType = DataType.VALUE
    inrange: bool = True
    temp_range: float = 0

    def __str__(self):
        return "name={} props={}, state={}".format(self.name, self.props, self.state)

    def to_dict(self):
        return dict(
            id=self.id,
            name=self.name,
            type=self.type,
            props=self.props.to_dict(),
            state=self.state,
        )


@dataclass
class Kettle:
    id: str = None
    name: str = None
    props: Props = Props()
    instance: str = None
    agitator: Actor = None
    heater: Actor = None
    sensor: Sensor = None
    type: str = None
    target_temp: float = 0

    def __str__(self):
        return "name={} props={} temp={}".format(
            self.name, self.props, self.target_temp
        )

    def to_dict(self):

        if self.instance is not None:

            state = self.instance.state

        else:
            state = False
        return dict(
            id=self.id,
            name=self.name,
            state=state,
            target_temp=self.target_temp,
            heater=self.heater,
            agitator=self.agitator,
            sensor=self.sensor,
            type=self.type,
            props=self.props.to_dict(),
        )


@dataclass
class Step:
    id: str = None
    name: str = None
    props: Props = Props()
    type: str = None
    status: StepState = StepState.INITIAL
    instance: str = None

    def __str__(self):
        return "name={} props={}, type={}, instance={}".format(
            self.name, self.props, self.type, self.instance
        )

    def to_dict(self):

        msg = self.instance.summary if self.instance is not None else ""
        msg2 = (
            self.instance.summary2
            if ((self.instance is not None) and (self.instance.summary2 is not None))
            else None
        )
        return dict(
            id=self.id,
            name=self.name,
            state_text=msg,
            state_text2=msg2,
            type=self.type,
            status=self.status.value,
            props=self.props.to_dict(),
        )


@dataclass
class Fermenter:
    id: str = None
    name: str = None
    sensor: Sensor = None
    pressure_sensor: Sensor = None
    heater: Actor = None
    cooler: Actor = None
    valve: Actor = None
    brewname: str = None
    description: str = None
    props: Props = Props()
    target_temp: float = 0
    target_pressure: float = 0
    type: str = None
    steps: List[Step] = field(default_factory=list)
    instance: str = None

    def __str__(self):
        return "name={} props={} temp={}".format(
            self.name, self.props, self.target_temp, self.target_pressure
        )

    def to_dict(self):

        if self.instance is not None:

            state = self.instance.state

        else:
            state = False

        steps = list(map(lambda item: item.to_dict(), self.steps))
        return dict(
            id=self.id,
            name=self.name,
            state=state,
            sensor=self.sensor,
            pressure_sensor=self.pressure_sensor,
            heater=self.heater,
            cooler=self.cooler,
            valve=self.valve,
            brewname=self.brewname,
            description=self.description,
            props=self.props.to_dict() if self.props is not None else None,
            target_temp=self.target_temp,
            target_pressure=self.target_pressure,
            type=self.type,
            steps=steps,
        )


@dataclass
class FermenterStep:
    id: str = None
    name: str = None
    fermenter: Fermenter = None
    props: Props = Props()
    type: str = None
    status: StepState = StepState.INITIAL
    endtime: int = 0  # endtime if step is active and timer is running
    instance: str = None
    step: dict = None

    def __str__(self):
        return "name={} props={}, type={}, instance={}".format(
            self.name, self.props, self.type, self.instance
        )

    def to_dict(self):
        msg = self.instance.summary if self.instance is not None else ""
        return dict(
            id=self.id,
            name=self.name,
            state_text=msg,
            type=self.type,
            status=self.status.value,
            endtime=self.endtime,
            props=self.props.to_dict(),
        )


class ConfigType(Enum):
    STRING = "string"
    ACTOR = "actor"
    SENSOR = "sensor"
    KETTLE = "kettle"
    NUMBER = "number"
    SELECT = "select"
    STEP = "step"
    FERMENTER = "fermenter"


@dataclass
class Config:

    name: str = None
    value: Any = None
    description: str = None
    type: ConfigType = ConfigType.STRING
    source: str = None
    options: Any = None

    def __str__(self):
        return "....name={} value={}".format(self.name, self.value)

    def to_dict(self):
        return dict(
            name=self.name,
            value=self.value,
            type=self.type.value,
            description=self.description,
            source=self.source,
            options=self.options,
        )


@dataclass
class NotificationAction:
    label: str
    method: Any = None
    id: str = None

    def to_dict(self):
        return dict(id=self.id, label=self.label)


class NotificationType(Enum):
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    SUCCESS = "success"

    def __str__(self):
        return self.value



# -*- coding: utf-8 -*-
import asyncio
import json
import logging
import time
from datetime import datetime

from cbpi.api import *
from cbpi.api import CBPiSensor, Property, parameters
from cbpi.api.dataclasses import NotificationAction, NotificationType


@parameters(
    [
        Property.Text(label="Topic", configurable=True, description="MQTT Topic"),
        Property.Text(
            label="PayloadDictionary",
            configurable=True,
            default_value="",
            description="Where to find msg in payload, leave blank for raw payload",
        ),
        Property.Kettle(
            label="Kettle",
            description="Reduced logging if Kettle is inactive / range warning in dashboard (only Kettle or Fermenter to be selected)",
        ),
        Property.Fermenter(
            label="Fermenter",
            description="Reduced logging if Fermenter is inactive / range warning in dashboard (only Kettle or Fermenter to be selected)",
        ),
        Property.Number(
            label="ReducedLogging",
            configurable=True,
            description="Reduced logging frequency in seconds if selected Kettle or Fermenter is inactive (default:60 sec | 0 disabled)",
        ),
        Property.Number(
            label="Timeout",
            configurable=True,
            unit="sec",
            description="Timeout in seconds to send notification (default:60 | deactivated: 0)",
        ),
        Property.Number(
            label="TempRange",
            configurable=True,
            unit="degree",
            description="Temp range in degree between reading and target temp of fermenter/kettle. Larger difference shows different color in dashboard (default:0 | deactivated: 0)",
        ),
    ]
)
class MQTTSensor(CBPiSensor):

    def __init__(self, cbpi, id, props):
        super(MQTTSensor, self).__init__(cbpi, id, props)
        self.Topic = self.props.get("Topic", None)
        self.payload_text = self.props.get("PayloadDictionary", None)
        if self.payload_text != None:
            self.payload_text = self.payload_text.split(".")
        self.subscribed = self.cbpi.satellite.subscribe(self.Topic, self.on_message)
        self.value: float = 999
        self.timeout = int(self.props.get("Timeout", 60))
        self.temprange = float(self.props.get("TempRange", 0))
        self.starttime = time.time()
        self.notificationsend = False
        self.nextchecktime = self.starttime + self.timeout
        self.lastdata = time.time()
        self.lastlog = 0
        self.sensor = self.get_sensor(self.id)
        self.reducedfrequency = int(self.props.get("ReducedLogging", 60))
        if self.reducedfrequency < 0:
            self.reducedfrequency = 0
        self.kettleid = self.props.get("Kettle", None)
        self.fermenterid = self.props.get("Fermenter", None)
        self.reducedlogging = True if self.kettleid or self.fermenterid else False

        if self.kettleid is not None and self.fermenterid is not None:
            self.reducedlogging = False
            self.cbpi.notify(
                "MQTTSensor",
                "Sensor '"
                + str(self.sensor.name)
                + "' cant't have Fermenter and Kettle defined for reduced logging / range warning.",
                NotificationType.WARNING,
                action=[NotificationAction("OK", self.Confirm)],
            )

    async def Confirm(self, **kwargs):
        self.nextchecktime = time.time() + self.timeout
        self.notificationsend = False
        pass

    async def message(self):
        target_timestring = datetime.fromtimestamp(self.lastdata)
        self.cbpi.notify(
            "MQTTSensor Timeout",
            "Sensor '"
            + str(self.sensor.name)
            + "' did not respond. Last data received: "
            + target_timestring.strftime("%D %H:%M"),
            NotificationType.WARNING,
            action=[NotificationAction("OK", self.Confirm)],
        )
        pass

    async def on_message(self, message):
        val = json.loads(message.payload.decode())
        try:
            if self.payload_text is not None:
                for key in self.payload_text:
                    val = val.get(key, None)

            if isinstance(val, (int, float, str)):
                self.value = float(val)
                self.push_update(self.value)
                if self.reducedlogging == True:
                    await self.logvalue()
                else:
                    logging.info(
                        "MQTTSensor {} regular logging".format(self.sensor.name)
                    )
                    self.log_data(self.value)
                    self.lastlog = time.time()

                if self.timeout != 0:
                    self.nextchecktime = time.time() + self.timeout
                    self.notificationsend = False
                    self.lastdata = time.time()
        except Exception as e:
            logging.error("MQTT Sensor Error {}".format(e))

    async def logvalue(self):
        self.kettle = (
            self.get_kettle(self.kettleid) if self.kettleid is not None else None
        )
        self.fermenter = (
            self.get_fermenter(self.fermenterid)
            if self.fermenterid is not None
            else None
        )
        now = time.time()
        if self.kettle is not None:
            try:
                kettlestatus = self.kettle.instance.state
            except:
                kettlestatus = False
            if kettlestatus:
                self.log_data(self.value)
                logging.info("MQTTSensor {} Kettle Active".format(self.sensor.name))
                self.lastlog = time.time()
            else:
                logging.info("MQTTSensor {} Kettle Inactive".format(self.sensor.name))
                if self.reducedfrequency != 0:
                    if now >= self.lastlog + self.reducedfrequency:
                        self.log_data(self.value)
                        self.lastlog = time.time()
                        logging.info("Logged with reduced freqency")
                        pass

        if self.fermenter is not None:
            try:
                fermenterstatus = self.fermenter.instance.state
            except:
                fermenterstatus = False
            if fermenterstatus:
                self.log_data(self.value)
                logging.info("MQTTSensor {} Fermenter Active".format(self.sensor.name))
                self.lastlog = time.time()
            else:
                logging.info(
                    "MQTTSensor {} Fermenter Inactive".format(self.sensor.name)
                )
                if self.reducedfrequency != 0:
                    if now >= self.lastlog + self.reducedfrequency:
                        self.log_data(self.value)
                        self.lastlog = time.time()
                        logging.info("Logged with reduced freqency")
                        pass

    async def run(self):
        while self.running:
            if self.timeout != 0:
                if time.time() > self.nextchecktime and self.notificationsend == False:
                    await self.message()
                    self.notificationsend = True
            await asyncio.sleep(1)

    def get_state(self):
        return dict(value=self.value)

    async def on_stop(self):
        self.subscribed = self.cbpi.satellite.unsubscribe(self.Topic, self.on_message)


@parameters(
    [
        Property.Text(label="Topic", configurable=True, description="MQTT Topic"),
        Property.Text(
            label="PayloadDictionary",
            configurable=True,
            default_value="",
            description="Where to find msg in payload, leave blank for raw payload",
        ),
        Property.Kettle(
            label="Kettle",
            description="Reduced logging if Kettle is inactive / range warning in dashboard (only Kettle or Fermenter to be selected)",
        ),
        Property.Fermenter(
            label="Fermenter",
            description="Reduced logging if Fermenter is inactive / range warning in dashboard (only Kettle or Fermenter to be selected)",
        ),
        Property.Number(
            label="Offset",
            configurable=True,
            description="Offset for MQTT Sensor (default is 0). !!! Use this only with caution as offset for MQTT sensor should be defined on Sensor side !!!",
        ),
        Property.Number(
            label="ReducedLogging",
            configurable=True,
            description="Reduced logging frequency in seconds if selected Kettle or Fermenter is inactive (default:60 sec | 0 disabled)",
        ),
        Property.Number(
            label="Timeout",
            configurable=True,
            unit="sec",
            description="Timeout in seconds to send notification (default:60 | deactivated: 0)",
        ),
        Property.Number(
            label="TempRange",
            configurable=True,
            unit="degree",
            description="Temp range in degree between reading and target temp of fermenter/kettle. Larger difference shows different color in dashboard (default:0 | deactivated: 0)",
        ),
    ]
)
class MQTTSensorOffset(CBPiSensor):

    def __init__(self, cbpi, id, props):
        super(MQTTSensorOffset, self).__init__(cbpi, id, props)
        self.Topic = self.props.get("Topic", None)
        self.offset = float(self.props.get("Offset", 0))
        self.payload_text = self.props.get("PayloadDictionary", None)
        if self.payload_text != None:
            self.payload_text = self.payload_text.split(".")
        self.subscribed = self.cbpi.satellite.subscribe(self.Topic, self.on_message)
        self.value: float = 999
        self.timeout = int(self.props.get("Timeout", 60))
        self.temprange = float(self.props.get("TempRange", 0))
        self.starttime = time.time()
        self.notificationsend = False
        self.nextchecktime = self.starttime + self.timeout
        self.lastdata = time.time()
        self.lastlog = 0
        self.sensor = self.get_sensor(self.id)
        self.reducedfrequency = int(self.props.get("ReducedLogging", 60))
        if self.reducedfrequency < 0:
            self.reducedfrequency = 0
        self.kettleid = self.props.get("Kettle", None)
        self.fermenterid = self.props.get("Fermenter", None)
        self.reducedlogging = True if self.kettleid or self.fermenterid else False

        if self.kettleid is not None and self.fermenterid is not None:
            self.reducedlogging = False
            self.cbpi.notify(
                "MQTTSensor",
                "Sensor '"
                + str(self.sensor.name)
                + "' cant't have Fermenter and Kettle defined for reduced logging / range warning.",
                NotificationType.WARNING,
                action=[NotificationAction("OK", self.Confirm)],
            )

    async def Confirm(self, **kwargs):
        self.nextchecktime = time.time() + self.timeout
        self.notificationsend = False
        pass

    async def message(self):
        target_timestring = datetime.fromtimestamp(self.lastdata)
        self.cbpi.notify(
            "MQTTSensor Timeout",
            "Sensor '"
            + str(self.sensor.name)
            + "' did not respond. Last data received: "
            + target_timestring.strftime("%D %H:%M"),
            NotificationType.WARNING,
            action=[NotificationAction("OK", self.Confirm)],
        )
        pass

    async def on_message(self, message):
        val = json.loads(message.payload.decode())
        try:
            if self.payload_text is not None:
                for key in self.payload_text:
                    val = val.get(key, None)

            if isinstance(val, (int, float, str)):
                self.value = float(val) + self.offset
                self.push_update(self.value)
                if self.reducedlogging == True:
                    await self.logvalue()
                else:
                    logging.info(
                        "MQTTSensor {} regular logging".format(self.sensor.name)
                    )
                    self.log_data(self.value)
                    self.lastlog = time.time()

                if self.timeout != 0:
                    self.nextchecktime = time.time() + self.timeout
                    self.notificationsend = False
                    self.lastdata = time.time()
        except Exception as e:
            logging.error("MQTT Sensor Error {}".format(e))

    async def logvalue(self):
        self.kettle = (
            self.get_kettle(self.kettleid) if self.kettleid is not None else None
        )
        self.fermenter = (
            self.get_fermenter(self.fermenterid)
            if self.fermenterid is not None
            else None
        )
        now = time.time()
        if self.kettle is not None:
            try:
                kettlestatus = self.kettle.instance.state
            except:
                kettlestatus = False
            if kettlestatus:
                self.log_data(self.value)
                logging.info("MQTTSensor {} Kettle Active".format(self.sensor.name))
                self.lastlog = time.time()
            else:
                logging.info("MQTTSensor {} Kettle Inactive".format(self.sensor.name))
                if self.reducedfrequency != 0:
                    if now >= self.lastlog + self.reducedfrequency:
                        self.log_data(self.value)
                        self.lastlog = time.time()
                        logging.info("Logged with reduced freqency")
                        pass

        if self.fermenter is not None:
            try:
                fermenterstatus = self.fermenter.instance.state
            except:
                fermenterstatus = False
            if fermenterstatus:
                self.log_data(self.value)
                logging.info("MQTTSensor {} Fermenter Active".format(self.sensor.name))
                self.lastlog = time.time()
            else:
                logging.info(
                    "MQTTSensor {} Fermenter Inactive".format(self.sensor.name)
                )
                if self.reducedfrequency != 0:
                    if now >= self.lastlog + self.reducedfrequency:
                        self.log_data(self.value)
                        self.lastlog = time.time()
                        logging.info("Logged with reduced freqency")
                        pass

    async def run(self):
        while self.running:
            if self.timeout != 0:
                if time.time() > self.nextchecktime and self.notificationsend == False:
                    await self.message()
                    self.notificationsend = True
            await asyncio.sleep(1)

    def get_state(self):
        return dict(value=self.value)

    async def on_stop(self):
        self.subscribed = self.cbpi.satellite.unsubscribe(self.Topic, self.on_message)


def setup(cbpi):
    """
    This method is called by the server during startup
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core
    :return:
    """
    if str(cbpi.static_config.get("mqtt", False)).lower() == "true":
        if str(cbpi.static_config.get("mqtt_offset", False)).lower() == "false":
            cbpi.plugin.register("MQTTSensor", MQTTSensor)
        else:
            cbpi.plugin.register("MQTTSensor", MQTTSensorOffset)



from cbpi.api import Property, parameters

from . import MQTTActor


@parameters(
    [
        Property.Text(label="Topic", configurable=True, description="MQTT Topic"),
        Property.Text(
            label="Payload",
            configurable=True,
            description="Payload that is sent as MQTT message. Available placeholders are {switch_onoff}: [on|off], {switch_10}: [1|0], {power}: [0-100].",
        ),
    ]
)
class GenericMqttActor(MQTTActor):
    def __init__(self, cbpi, id, props):
        MQTTActor.__init__(self, cbpi, id, props)
        self.payload = ""

    async def on_start(self):
        await MQTTActor.on_start(self)
        self.payload = self.props.get(
            "Payload", '{{"state": "{switch_onoff}", "power": {power}}}'
        )

    def normalize_power_value(self, power):
        if power is not None:
            if power != self.power:
                power = min(100, power)
                power = max(0, power)
                self.power = round(power)

    async def publish_mqtt_message(self, topic, payload):
        self.logger.info(
            "Publish '{payload}' to '{topic}'".format(payload=payload, topic=self.topic)
        )
        await self.cbpi.satellite.publish(self.topic, payload, True)

    async def on(self, power=None):
        self.normalize_power_value(power)
        formatted_payload = self.payload.format(
            switch_onoff="on", switch_10=1, power=self.power
        )
        await self.publish_mqtt_message(self.topic, formatted_payload)
        self.state = True

    async def off(self):
        formatted_payload = self.payload.format(
            switch_onoff="off", switch_10=0, power=0
        )
        await self.publish_mqtt_message(self.topic, formatted_payload)
        self.state = False
